from formalgeo.tools import load_json, save_json, get_user_input
import os
import requests
from tqdm import tqdm
import json
import tarfile
import shutil
import random

remote = "https://raw.githubusercontent.com/Formalgeo/Datasets/main/released/"


def get_datasets_path(datasets_path):
    if datasets_path is None:
        datasets_path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "datasets")
    if not os.path.exists(datasets_path):
        os.makedirs(datasets_path)
    return datasets_path


def format_json(path_datasets):
    for path, _, files in os.walk(path_datasets):
        for file in files:
            if file.endswith(".json"):  # format json
                save_json(load_json(os.path.join(path, file)), os.path.join(path, file))


def get_remote_datasets():
    response = requests.get(os.path.join(remote, "released.json"))
    if response.status_code == 200:
        return json.loads(response.content)
    return None


def get_local_datasets(datasets_path):
    local_datasets = {}
    for file in os.listdir(datasets_path):
        if file.endswith(".json"):
            info = load_json(os.path.join(datasets_path, file))
            local_datasets[file.split(".")[0]] = info

    return local_datasets


def show_available_datasets(datasets_path=None):
    datasets_path = get_datasets_path(datasets_path)
    remote_datasets = get_remote_datasets()
    local_datasets = get_local_datasets(datasets_path)
    text = "{0:<15}{1:<15}{2:<15}{3:<15}{4:<15}{5:<15}{6:<22}{7:}"
    print(text.format("Location", "Name", "Version", "Formalgeo", "GDL", "GDL-Version", "Release", "Description"))

    if remote_datasets is None:
        for dataset in local_datasets:
            print(text.format("local",
                              local_datasets[dataset]["dataset_name"],
                              local_datasets[dataset]["dataset_version"],
                              local_datasets[dataset]["formalgeo_version"],
                              local_datasets[dataset]["gdl_name"],
                              local_datasets[dataset]["gdl_version"],
                              local_datasets[dataset]["release_datetime"],
                              local_datasets[dataset]["short_description"]))
        print("\nFailed to get the remote datasets, displaying local datasets.")
        print("Please check your network connection and try again.")
        return

    for dataset in local_datasets:
        if dataset in remote_datasets:
            if local_datasets[dataset]["release_datetime"] == remote_datasets[dataset]["release_datetime"]:
                dataset_name = "\033[32m{}\033[0m".format(local_datasets[dataset]["dataset_name"])
            else:
                dataset_name = "\033[33m{}\033[0m".format(local_datasets[dataset]["dataset_name"])
            print(text.format("local",
                              dataset_name,
                              local_datasets[dataset]["dataset_version"],
                              local_datasets[dataset]["formalgeo_version"],
                              local_datasets[dataset]["gdl_name"],
                              local_datasets[dataset]["gdl_version"],
                              local_datasets[dataset]["release_datetime"],
                              local_datasets[dataset]["short_description"]))
            print(text.format("remote",
                              remote_datasets[dataset]["dataset_name"],
                              remote_datasets[dataset]["dataset_version"],
                              remote_datasets[dataset]["formalgeo_version"],
                              remote_datasets[dataset]["gdl_name"],
                              remote_datasets[dataset]["gdl_version"],
                              remote_datasets[dataset]["release_datetime"],
                              remote_datasets[dataset]["short_description"]))
        else:
            print(text.format("\033[34m{}\033[0m".format(local_datasets[dataset]["dataset_name"]),
                              local_datasets[dataset]["dataset_version"],
                              local_datasets[dataset]["formalgeo_version"],
                              local_datasets[dataset]["gdl_name"],
                              local_datasets[dataset]["gdl_version"],
                              local_datasets[dataset]["release_datetime"],
                              local_datasets[dataset]["short_description"]))

    for dataset in remote_datasets:
        if dataset not in local_datasets:
            print(text.format("remote",
                              remote_datasets[dataset]["dataset_name"],
                              remote_datasets[dataset]["dataset_version"],
                              remote_datasets[dataset]["formalgeo_version"],
                              remote_datasets[dataset]["gdl_name"],
                              remote_datasets[dataset]["gdl_version"],
                              remote_datasets[dataset]["release_datetime"],
                              remote_datasets[dataset]["short_description"]))

    print("\n<\033[32mGreen\033[0m> indicates that the local version is in sync with the remote version.")
    print("<\033[33mYellow\033[0m> indicates the remote dataset has updates.")
    print("<\033[34mBlue\033[0m> represents the local dataset not in the remote repository.")
    print("<White> represents the remote dataset that can be downloaded.")
    print("Run 'formalgeo.data.download_dataset()' to download the remote dataset.")


def download_dataset(dataset_name, datasets_path=None):
    datasets_path = get_datasets_path(datasets_path)
    remote_datasets = get_remote_datasets()
    local_datasets = get_local_datasets(datasets_path)

    if remote_datasets is None:
        msg = "Network error. Fail to get remote datasets lists."
        raise Exception(msg)

    if dataset_name not in remote_datasets:
        msg = "No dataset named '{}'. run 'formalgeo.data.show_available_datasets()' for more info.".format(
            dataset_name)
        raise Exception(msg)

    if dataset_name in local_datasets:
        if local_datasets[dataset_name]["release_datetime"] == remote_datasets[dataset_name]["release_datetime"]:
            print("Datasets '{}' already exits in '{}'.".format(dataset_name, datasets_path))
            if get_user_input("Do you want to update?") == "n":
                return False
        remove_dataset(dataset_name, datasets_path)

    response = requests.get(os.path.join(remote, "{}.tar.gz".format(dataset_name)), stream=True)
    if response.status_code == 200:
        pbar = tqdm(
            total=int(response.headers.get('content-length', 0)),
            unit='iB',
            unit_scale=True,
            desc="Download dataset '{}'".format(dataset_name)
        )
        with open(os.path.join(datasets_path, "{}.tar.gz".format(dataset_name)), "wb") as file:
            for data in response.iter_content(1024):  # block_size = 1024
                pbar.update(len(data))
                file.write(data)
        pbar.close()

        print("Extracting files ...")
        with tarfile.open(os.path.join(datasets_path, "{}.tar.gz".format(dataset_name)), "r:gz") as tar_file:  # extract
            tar_file.extractall(os.path.join(datasets_path, dataset_name))

        format_json(os.path.join(datasets_path, dataset_name))

        shutil.copy(os.path.join(datasets_path, dataset_name, "info.json"),
                    os.path.join(datasets_path, "{}.json".format(dataset_name)))

        os.remove(os.path.join(datasets_path, "{}.tar.gz".format(dataset_name)))

        return True

    msg = "Network error. Fail to download '{}'.".format(
        os.path.join(remote, os.path.join(remote, "{}.tar.gz".format(dataset_name))))
    raise Exception(msg)


def remove_dataset(dataset_name, datasets_path=None):
    datasets_path = get_datasets_path(datasets_path)
    local_datasets = get_local_datasets(datasets_path)

    if dataset_name not in local_datasets:
        print("No dataset '{}' in '{}'. please check your inputs.".format(dataset_name, datasets_path))
        return

    print("Removing dataset '{}'...".format(dataset_name))
    if os.path.isdir(os.path.join(datasets_path, dataset_name)):
        shutil.rmtree(os.path.join(datasets_path, dataset_name), ignore_errors=True)


class DatasetLoader:

    def __init__(self, dataset_name, datasets_path=None):
        datasets_path = get_datasets_path(datasets_path)
        local_datasets = get_local_datasets(datasets_path)
        if dataset_name not in local_datasets:
            msg = "No dataset dir named '{}'. run 'formalgeo.data.show_available_datasets()' for more info.".format(
                dataset_name)
            raise Exception(msg)

        self.dataset_path = os.path.join(datasets_path, dataset_name)

        self.info = load_json(os.path.join(self.dataset_path, "info.json"))
        self.predicate_GDL = load_json(os.path.join(self.dataset_path, "gdl", "predicate_GDL.json"))
        self.theorem_GDL = load_json(os.path.join(self.dataset_path, "gdl", "theorem_GDL.json"))
        self.pid_mapping = self.load_file("pid_mapping.json")

    def show(self):
        for item in self.info:
            print("{}: {}".format(item, self.info[item]))
        print("dataset_path: {}".format(self.dataset_path))
        print("files: {}".format(os.listdir(os.path.join(self.dataset_path, "files"))))

    def get_problem(self, pid):
        if pid <= self.info["problem_number"]:
            return load_json(os.path.join(self.dataset_path, "problems", "{}.json".format(pid)))
        elif pid <= self.info["expanded_problem_number"]:
            raw_pid = self.pid_mapping[str(pid)]
            raw_problem = load_json(os.path.join(self.dataset_path, "problems", "{}.json".format(raw_pid)))
            expanded = load_json(os.path.join(self.dataset_path, "expanded", "{}.json".format(raw_pid)))[str(pid)]
            raw_problem["problem_id"] = expanded["problem_id"]
            raw_problem["text_cdl"] += expanded["added_cdl"]
            raw_problem["goal_cdl"] = expanded["goal_cdl"]
            raw_problem["problem_answer"] = expanded["problem_answer"]
            raw_problem["theorem_seqs"] = expanded["theorem_seqs"]
            raw_problem["theorem_seqs_dag"] = expanded["theorem_seqs_dag"]
            raw_problem["problem_level"] = len(expanded["theorem_seqs"])
            return raw_problem
        else:
            msg = "No problem named {}.".format(pid)
            raise Exception(msg)

    def get_problem_split(self, expanded=False):
        file_path = os.path.join(self.dataset_path, "files")
        if expanded:
            split_msg = self.info["expanded_problem_split"]
            filename = "expanded_problem_split_{}_{}_{}_{}.json".format(
                split_msg[0], split_msg[1], split_msg[2], split_msg[3]
            )
            problem_number = self.info["expanded_problem_number"]
        else:
            split_msg = self.info["problem_split"]
            filename = "problem_split_{}_{}_{}_{}.json".format(
                split_msg[0], split_msg[1], split_msg[2], split_msg[3]
            )
            problem_number = self.info["problem_number"]

        if filename in os.listdir(file_path):
            return load_json(os.path.join(file_path, filename))

        total = split_msg[0] + split_msg[1] + split_msg[2]
        random.seed(split_msg[3])
        data = list(range(1, problem_number + 1))
        test = sorted(random.sample(data, int(problem_number * split_msg[2] / total)))
        for i in range(len(data))[::-1]:
            if data[i] in test:
                data.pop(i)
        val = sorted(random.sample(data, int(problem_number * split_msg[1] / total)))
        for i in range(len(data))[::-1]:
            if data[i] in val:
                data.pop(i)
        train = data

        total = len(train) + len(val) + len(test)

        data = {
            "msg": {
                "total": total,
                "train": len(train),
                "val": len(val),
                "test": len(test)
            },
            "split": {
                "train": train,
                "val": val,
                "test": test
            }
        }

        save_json(data, os.path.join(file_path, filename))

        return data

    def load_file(self, filename):
        file_path = os.path.join(self.dataset_path, "files")
        if filename not in os.listdir(file_path):
            msg = "No file named {} in {}.".format(filename, file_path)
            raise Exception(msg)

        if filename.endswith(".json"):
            return load_json(os.path.join(file_path, filename))
        else:
            with open(os.path.join(file_path, filename)) as f:
                return f.read()



# Copyright (C) 2022-2023 FormalGeo Development Team
# Author: Xiaokai Zhang
# Contact: formalgeo@gmail.com

"""Formal representation and solving for Euclidean plane geometry problems."""

__all__ = [
    "core", "problem", "solver", "tools"
]



from formalgeo.problem import Problem
from formalgeo.core import GeometryPredicateLogicExecutor as GPLExecutor
from formalgeo.core import EquationKiller as EqKiller
from formalgeo.parse import parse_predicate_gdl, parse_theorem_gdl, parse_problem_cdl
from formalgeo.parse import get_equation_from_tree
from formalgeo.tools import rough_equal
import warnings
import time


class Interactor:

    def __init__(self, predicate_GDL, theorem_GDL):
        """
        Initialize Interactor.
        :param predicate_GDL: predicate GDL.
        :param theorem_GDL: theorem GDL.
        """
        self.parsed_predicate_GDL = parse_predicate_gdl(predicate_GDL)
        self.parsed_theorem_GDL = parse_theorem_gdl(theorem_GDL, self.parsed_predicate_GDL)
        self.problem = None

    def load_problem(self, problem_CDL):
        """Load problem through problem_CDL."""
        start_time = time.time()
        self.problem = Problem()
        self.problem.load_problem_by_fl(self.parsed_predicate_GDL,
                                        self.parsed_theorem_GDL,
                                        parse_problem_cdl(problem_CDL))  # load problem
        EqKiller.solve_equations(self.problem)  # Solve the equations after initialization
        self.problem.step("init_problem", time.time() - start_time)  # save applied theorem and update step

    def apply_theorem(self, t_name, t_branch=None, t_para=None):
        """
        Apply a theorem and return whether it is successfully applied.
        :param t_name: <str>.
        :param t_branch: <str>.
        :param t_para: tuple of <str>.
        :return update: <bool>, Whether the question condition updated or not.
        """
        if self.problem is None:
            e_msg = "Problem not loaded. Please run <load_problem> before run <apply_theorem>."
            raise Exception(e_msg)
        if t_name not in self.parsed_theorem_GDL:
            e_msg = "Theorem {} not defined in current GDL.".format(t_name)
            raise Exception(e_msg)
        if t_name.endswith("definition"):
            e_msg = "Theorem {} only used for backward reason.".format(t_name)
            raise Exception(e_msg)
        if t_para is not None and len(t_para) != len(self.parsed_theorem_GDL[t_name]["vars"]):
            e_msg = "Theorem <{}> para length error. Expected {} but got {}.".format(
                t_name, len(self.parsed_theorem_GDL[t_name]["vars"]), t_para)
            raise Exception(e_msg)
        if t_branch is not None and t_branch not in self.parsed_theorem_GDL[t_name]["body"]:
            e_msg = "Theorem <{}> branch error. Expected {} but got {}.".format(
                t_name, self.parsed_theorem_GDL[t_name]["body"].keys(), t_branch)
            raise Exception(e_msg)

        if t_para is None and t_branch is None:
            update = self.apply_theorem_by_name(t_name)
        elif t_para is not None and t_branch is None:
            update = self.apply_theorem_by_name_and_para(t_name, t_para)
        elif t_para is None and t_branch is not None:
            update = self.apply_theorem_by_name_and_branch(t_name, t_branch)
        else:
            update = self.apply_theorem_by_name_and_para_and_branch(t_name, t_branch, t_para)

        if not update:
            w_msg = "Theorem <{},{},{}> not applied. Please check your theorem or prerequisite.".format(
                t_name, t_branch, t_para)
            warnings.warn(w_msg)

        return update

    def apply_theorem_by_name(self, t_name):
        """
        Apply a theorem with t_name.
        :param t_name: <str>.
        :return update: <bool>, Whether the problem condition updated or not.
        """
        update = False

        for branch in self.parsed_theorem_GDL[t_name]["body"]:
            timing = time.time()  # timing
            gpl = self.parsed_theorem_GDL[t_name]["body"][branch]

            conclusions = GPLExecutor.run(gpl, self.problem)  # get gpl reasoned result
            if len(conclusions) == 0:
                theorem = (t_name, branch, None)
                self.problem.step(theorem, time.time() - timing)
                continue
            avg_timing = (time.time() - timing) / len(conclusions)
            for letters, premise, conclusion in conclusions:
                t_para = [letters[i] for i in self.parsed_theorem_GDL[t_name]["vars"]]
                theorem = (t_name, branch, tuple(t_para))
                for predicate, item in conclusion:  # add conclusion
                    update = self.problem.add(predicate, item, premise, theorem) or update
                self.problem.step(theorem, avg_timing)

        timing = time.time()  # timing
        EqKiller.solve_equations(self.problem)
        self.problem.step("solve_equations", time.time() - timing)

        return update

    def apply_theorem_by_name_and_para(self, t_name, t_para):
        """
        Apply a theorem with t_name and t_para.
        :param t_name: <str>.
        :param t_para: tuple of <str>.
        :return update: <bool>, Whether the problem condition updated or not.
        """
        update = False
        letters = {}  # used for vars-letters replacement
        for i in range(len(self.parsed_theorem_GDL[t_name]["vars"])):
            letters[self.parsed_theorem_GDL[t_name]["vars"][i]] = t_para[i]

        for branch in self.parsed_theorem_GDL[t_name]["body"]:
            timing = time.time()  # timing
            theorem = (t_name, branch, t_para)
            gpl = self.parsed_theorem_GDL[t_name]["body"][branch]
            premises = []
            passed = True

            for predicate, item in gpl["products"] + gpl["logic_constraints"]:
                oppose = False
                if "~" in predicate:
                    oppose = True
                    predicate = predicate.replace("~", "")
                item = tuple(letters[i] for i in item)
                has_item = self.problem.condition.has(predicate, item)
                if has_item:
                    premises.append(self.problem.condition.get_id_by_predicate_and_item(predicate, item))

                if (not oppose and not has_item) or (oppose and has_item):
                    passed = False
                    break

            if not passed:
                self.problem.step(theorem, time.time() - timing)
                continue

            for equal, item in gpl["algebra_constraints"]:
                oppose = False
                if "~" in equal:
                    oppose = True
                eq = get_equation_from_tree(self.problem, item, True, letters)
                solved_eq = False

                result, premise = EqKiller.solve_target(eq, self.problem)
                if result is not None and rough_equal(result, 0):
                    solved_eq = True
                premises += premise

                if (not oppose and not solved_eq) or (oppose and solved_eq):
                    passed = False
                    break

            if not passed:
                self.problem.step(theorem, time.time() - timing)
                continue

            for predicate, item in gpl["conclusions"]:
                if predicate == "Equal":  # algebra conclusion
                    eq = get_equation_from_tree(self.problem, item, True, letters)
                    update = self.problem.add("Equation", eq, premises, theorem) or update
                else:  # logic conclusion
                    item = tuple(letters[i] for i in item)
                    update = self.problem.add(predicate, item, premises, theorem) or update

            self.problem.step(theorem, time.time() - timing)

        timing = time.time()  # timing
        EqKiller.solve_equations(self.problem)
        self.problem.step("solve_equations", time.time() - timing)

        return update

    def apply_theorem_by_name_and_branch(self, t_name, t_branch):
        """
        Apply a theorem with t_name and t_branch.
        :param t_name: <str>.
        :param t_branch: <str>.
        :return update: <bool>, Whether the question condition updated or not.
        """
        update = False

        timing = time.time()  # timing
        gpl = self.parsed_theorem_GDL[t_name]["body"][t_branch]

        conclusions = GPLExecutor.run(gpl, self.problem)  # get gpl reasoned result
        if len(conclusions) == 0:
            theorem = (t_name, t_branch, None)
            self.problem.step(theorem, time.time() - timing)
            return False
        avg_timing = (time.time() - timing) / len(conclusions)
        for letters, premise, conclusion in conclusions:
            t_para = [letters[i] for i in self.parsed_theorem_GDL[t_name]["vars"]]
            theorem = (t_name, t_branch, tuple(t_para))

            for predicate, item in conclusion:  # add conclusion
                update = self.problem.add(predicate, item, premise, theorem) or update
            self.problem.step(theorem, avg_timing)

        timing = time.time()  # timing
        EqKiller.solve_equations(self.problem)
        self.problem.step("solve_equations", time.time() - timing)

        return update

    def apply_theorem_by_name_and_para_and_branch(self, t_name, t_branch, t_para):
        """
        Apply a theorem with t_name, t_branch and t_para.
        :param t_name: <str>.
        :param t_branch: <str>.
        :param t_para: tuple of <str>.
        :return update: <bool>, Whether the problem condition updated or not.
        """
        update = False
        timing = time.time()  # timing
        theorem = (t_name, t_branch, t_para)

        letters = {}  # used for vars-letters replacement
        for i in range(len(self.parsed_theorem_GDL[t_name]["vars"])):
            letters[self.parsed_theorem_GDL[t_name]["vars"][i]] = t_para[i]

        gpl = self.parsed_theorem_GDL[t_name]["body"][t_branch]
        premises = []

        for predicate, item in gpl["products"] + gpl["logic_constraints"]:
            oppose = False
            if "~" in predicate:
                oppose = True
                predicate = predicate.replace("~", "")
            item = tuple(letters[i] for i in item)
            has_item = self.problem.condition.has(predicate, item)
            if has_item:
                premises.append(self.problem.condition.get_id_by_predicate_and_item(predicate, item))

            if (not oppose and not has_item) or (oppose and has_item):
                self.problem.step(theorem, time.time() - timing)
                return False

        for equal, item in gpl["algebra_constraints"]:
            oppose = False
            if "~" in equal:
                oppose = True
            eq = get_equation_from_tree(self.problem, item, True, letters)
            solved_eq = False

            result, premise = EqKiller.solve_target(eq, self.problem)
            if result is not None and rough_equal(result, 0):
                solved_eq = True
            premises += premise

            if (not oppose and not solved_eq) or (oppose and solved_eq):
                self.problem.step(theorem, time.time() - timing)
                return False

        for predicate, item in gpl["conclusions"]:
            if predicate == "Equal":  # algebra conclusion
                eq = get_equation_from_tree(self.problem, item, True, letters)
                update = self.problem.add("Equation", eq, premises, theorem) or update
            else:  # logic conclusion
                item = tuple(letters[i] for i in item)
                update = self.problem.add(predicate, item, premises, theorem) or update
        self.problem.step(theorem, time.time() - timing)

        timing = time.time()  # timing
        EqKiller.solve_equations(self.problem)
        self.problem.step("solve_equations", time.time() - timing)

        return update



import time
import copy
import random
from enum import Enum
from itertools import permutations
from formalgeo.problem import Problem
from formalgeo.core import EquationKiller as EqKiller
from formalgeo.parse import parse_predicate_gdl, parse_theorem_gdl, parse_problem_cdl, get_equation_from_tree
from formalgeo.tools import get_used_pid_and_theorem, debug_print


def get_p2t_map_bw(t_info, parsed_theorem_GDL):
    """
    Predicate-theorem mapping hash table for Search Accelerating.
    :param t_info: <dict>, {t_name: (category_id, usage_count)}, user customization.
    :param parsed_theorem_GDL: parsed theorem GDL.
    """
    p2t_map_bw = {}
    for t_name in parsed_theorem_GDL:
        if t_name in t_info and (t_info[t_name][1] == 0 or t_info[t_name][0] == 3):  # skip no used and diff t
            continue
        for branch in parsed_theorem_GDL[t_name]["body"]:
            theorem_unit = parsed_theorem_GDL[t_name]["body"][branch]
            conclusions = list(theorem_unit["conclusions"])
            conclusions += list(theorem_unit["attr_in_conclusions"])
            for predicate, _ in conclusions:
                if predicate == "Equal":
                    continue
                if predicate not in p2t_map_bw:
                    p2t_map_bw[predicate] = [(t_name, branch)]
                elif (t_name, branch) not in p2t_map_bw[predicate]:
                    p2t_map_bw[predicate].append((t_name, branch))
    return p2t_map_bw


class GoalFinder:

    def __init__(self, parsed_theorem_GDL, p2t_map):
        self.parsed_theorem_GDL = parsed_theorem_GDL
        self.p2t_map = p2t_map

    def find_all_sub_goals(self, predicate, item, problem):
        """return [(sub_goals, (t_name, t_para, t_branch))]"""
        theorem_and_para = {}  # {(t_name, t_branch): set(t_paras)}

        if predicate == "Equation":  # algebra goal
            attr_to_paras = {}
            for sym in EqKiller.get_minimum_syms([item], list(problem.condition.simplified_equation)):
                attr, paras = problem.condition.attr_of_sym[sym]
                if attr == "Free":
                    continue
                if attr not in attr_to_paras:
                    attr_to_paras[attr] = []
                for para in paras:
                    if para not in attr_to_paras[attr]:
                        attr_to_paras[attr].append(para)

            for attr in attr_to_paras:
                for t_name, t_branch in self.p2t_map[attr]:
                    if (t_name, t_branch) not in theorem_and_para:
                        theorem_and_para[(t_name, t_branch)] = set()
                    t_paras = []
                    for t_attr, attr_vars in self.parsed_theorem_GDL[t_name]["body"][t_branch]["attr_in_conclusions"]:
                        if attr != t_attr:
                            continue
                        t_vars = copy.copy(self.parsed_theorem_GDL[t_name]["vars"])
                        for para in attr_to_paras[attr]:
                            t_para = [v if v not in attr_vars else para[attr_vars.index(v)] for v in t_vars]
                            t_paras.append(t_para)
                    t_paras = GoalFinder.theorem_para_completion(
                        t_paras, problem.condition.get_items_by_predicate("Point"))
                    theorem_and_para[(t_name, t_branch)] |= t_paras
        else:  # logic goal
            if predicate not in self.p2t_map:
                return []
            for t_name, t_branch in self.p2t_map[predicate]:
                if (t_name, t_branch) not in theorem_and_para:
                    theorem_and_para[(t_name, t_branch)] = set()
                t_paras = set()

                for t_predicate, item_vars in self.parsed_theorem_GDL[t_name]["body"][t_branch]["conclusions"]:
                    if t_predicate != predicate:
                        continue
                    t_vars = copy.copy(self.parsed_theorem_GDL[t_name]["vars"])
                    t_para = [v if v not in item_vars else item[item_vars.index(v)] for v in t_vars]
                    t_paras.add(tuple(t_para))

                t_paras = GoalFinder.theorem_para_completion(
                    t_paras, problem.condition.get_items_by_predicate("Point"))

                theorem_and_para[(t_name, t_branch)] |= t_paras

        return self.gen_sub_goals(theorem_and_para, problem)

    @staticmethod
    def theorem_para_completion(t_paras, points):
        """
        Replace free vars with points.
        >> theorem_para_completion([['a', 'R', 'S']], ['A', 'R', 'S'])
        >> [['A', 'R', 'S'], ['R', 'R', 'S'], ['S', 'R', 'S']]
        """
        points = [points[i][0] for i in range(len(points))]
        results = set()
        for t_para in t_paras:
            vacant_index = []
            for i in range(len(t_para)):
                if t_para[i].islower():
                    vacant_index.append(i)
            for per_para in permutations(points, len(vacant_index)):
                result = [t_para[i] if i not in vacant_index else per_para[vacant_index.index(i)]
                          for i in range(len(t_para))]
                results.add(tuple(result))
        return results

    def gen_sub_goals(self, theorem_and_para, problem):
        """
        Construct and return legitimate sub goal.
        :param theorem_and_para: {(t_name, t_branch): t_paras}.
        :param problem: Class <Problem>.
        :return results: [(t_name, t_branch, t_para, sub_goals)].
        """
        results = []
        for t in theorem_and_para:
            t_name, t_branch = t
            t_vars = self.parsed_theorem_GDL[t_name]["vars"]
            for t_para in theorem_and_para[t]:
                letters = {}
                for j in range(len(t_vars)):
                    letters[t_vars[j]] = t_para[j]

                sub_goals = []
                passed = True
                for predicate, item_vars in self.parsed_theorem_GDL[t_name]["body"][t_branch]["products"]:
                    item = tuple(letters[i] for i in item_vars)
                    if not (problem.ee_check(predicate, item) and problem.fv_check(predicate, item)):
                        passed = False
                        break
                    sub_goals.append((predicate, item))
                if not passed:
                    continue

                for predicate, item_vars in self.parsed_theorem_GDL[t_name]["body"][t_branch]["logic_constraints"]:
                    item = tuple(letters[i] for i in item_vars)
                    if not (problem.ee_check(predicate, item) and problem.fv_check(predicate, item)):
                        passed = False
                        break
                    sub_goals.append((predicate, item))
                if not passed:
                    continue

                for _, tree in self.parsed_theorem_GDL[t_name]["body"][t_branch]["algebra_constraints"]:
                    eq = get_equation_from_tree(problem, tree, True, letters)
                    if eq is None:
                        passed = False
                        break
                    sub_goals.append(("Equation", eq))
                if not passed:
                    continue

                result = (t_name, t_branch, t_para, tuple(sub_goals))
                if result not in results:
                    results.append(result)

        return results


class NodeState(Enum):
    to_be_expanded = 1
    expanded = 2
    success = 3
    fail = 4


class Node:
    def __init__(self, super_node, problem, predicate, item, node_map, finder, debug):
        """Init node and set node state."""
        self.state = NodeState.to_be_expanded
        self.super_node = super_node  # class <SuperNode>
        self.children = []  # list of class <SuperNode>
        self.children_t_msg = set()  # set of (t_name, t_branch, t_para)

        self.problem = problem
        self.predicate = predicate
        self.item = item
        self.premise = []

        self.finder = finder
        self.node_map = node_map

        self.debug = debug

        if predicate == "Equation":  # process 1
            for sym in self.item.free_symbols:
                if sym not in node_map:
                    node_map[sym] = [self]
                else:
                    node_map[sym].append(self)
            if item == 0:
                self.state = NodeState.success
        else:
            if (predicate, item) not in node_map:
                node_map[(predicate, item)] = [self]
            else:
                node_map[(predicate, item)].append(self)

            if predicate in ["Point", "Line", "Arc", "Angle", "Polygon", "Circle", "Collinear", "Cocircular"] and \
                    item not in self.problem.condition.get_items_by_predicate(predicate):
                self.state = NodeState.fail

        self.check_goal()

    def check_state(self):  # process 3
        if self.state in [NodeState.success, NodeState.fail]:
            return

        update = False
        fail = True
        success = False
        for child in self.children:
            if child.state != NodeState.fail:
                fail = False
            if child.state == NodeState.success:
                success = True

        if success:
            update = self.check_goal() or update

        if fail and len(self.children) > 0 and self.predicate != "Equation":
            self.state = NodeState.fail
            update = True

        if update:
            self.super_node.check_state()

    def check_goal(self):  # process 1
        """Return update or not"""
        if self.state in [NodeState.success, NodeState.fail]:
            return False

        if self.predicate == "Equation":
            result, premise = EqKiller.solve_target(self.item, self.problem)
            if result is None:
                return False

            if result == 0:
                self.state = NodeState.success
                self.premise = premise
            else:
                self.state = NodeState.fail
        else:
            if self.item not in self.problem.condition.get_items_by_predicate(self.predicate):
                return False
            self.state = NodeState.success
            self.premise = [self.problem.condition.get_id_by_predicate_and_item(self.predicate, self.item)]

        return True

    def expand(self, search_stack):  # process 1
        if self.state in [NodeState.success, NodeState.fail]:
            return False
        self.state = NodeState.expanded

        depth = self.super_node.pos[0] + 1
        results = self.finder.find_all_sub_goals(self.predicate, self.item, self.problem)
        for t_name, t_branch, t_para, sub_goals in results:
            if (t_name, t_branch, t_para) in self.children_t_msg:
                continue
            self.children_t_msg.add((t_name, t_branch, t_para))

            super_node = SuperNode(self, self.problem, (t_name, t_branch, t_para), depth,
                                   self.node_map, self.finder, self.debug, search_stack)
            self.children.append(super_node)
            super_node.add_nodes(sub_goals)


class SuperNode:
    snc = {}  # {depth: super_node_count}

    def __init__(self, father_node, problem, theorem, depth, node_map, finder, debug, search_stack):
        self.state = NodeState.to_be_expanded
        self.nodes = []  # list of class <Node>
        self.father_node = father_node  # class <Node>
        self.problem = problem
        self.theorem = theorem  # (t_name, t_branch, t_para)
        if depth not in SuperNode.snc:
            SuperNode.snc[depth] = 0
        self.pos = (depth, SuperNode.snc[depth] + 1)  # (depth, node_number)
        SuperNode.snc[depth] += 1
        self.node_map = node_map
        self.finder = finder
        self.debug = debug

        self.search_stack = search_stack
        search_stack.append(self)

    def add_nodes(self, sub_goals):
        father_super_nodes = []  # ensure no ring
        if self.father_node is not None:
            father_super_nodes.append(self.father_node.super_node)
        while len(father_super_nodes) > 0:
            super_node = father_super_nodes.pop()
            if super_node.theorem == self.theorem:
                self.state = NodeState.fail
                self.father_node.check_state()
                return
            if super_node.father_node is not None:
                father_super_nodes.append(super_node.father_node.super_node)

        for predicate, item in sub_goals:
            node = Node(self, self.problem, predicate, item, self.node_map, self.finder, self.debug)
            self.nodes.append(node)
            if node.state == NodeState.fail:
                break

        self.check_state()

    def check_state(self):  # process 2
        if self.state in [NodeState.success, NodeState.fail]:
            return

        for node in self.nodes:
            if node.state == NodeState.fail:
                self.state = NodeState.fail
                if self.father_node is not None:
                    self.father_node.check_state()
                return

        success = True
        for node in self.nodes:
            if node.state != NodeState.success:
                success = False
                break

        if success:
            self.state = NodeState.success
            if self.father_node is not None:
                self.apply_theorem()
                self.father_node.check_state()

    def expand(self):
        self.state = NodeState.expanded

        for i in range(len(self.nodes)):
            if self.state == NodeState.success:
                break
            debug_print(self.debug, "(pid={},depth={},branch={}/{},nodes={}/{}) Expanding Node ({}, {})".format(
                self.problem.parsed_problem_CDL["id"], self.pos[0], self.pos[1], SuperNode.snc[self.pos[0]],
                i + 1, len(self.nodes), self.nodes[i].predicate, self.nodes[i].item))
            self.nodes[i].expand(self.search_stack)

    def apply_theorem(self):
        if self.theorem is None or self.theorem[0].endswith("definition"):
            return

        t_name, t_branch, t_para = self.theorem

        letters = {}  # used for vars-letters replacement
        for i in range(len(self.finder.parsed_theorem_GDL[t_name]["vars"])):
            letters[self.finder.parsed_theorem_GDL[t_name]["vars"][i]] = t_para[i]

        gpl = self.finder.parsed_theorem_GDL[t_name]["body"][t_branch]
        premises = []
        passed = True

        for predicate, item in gpl["products"] + gpl["logic_constraints"]:
            oppose = False
            if "~" in predicate:
                oppose = True
                predicate = predicate.replace("~", "")
            item = tuple(letters[i] for i in item)
            has_item = self.problem.condition.has(predicate, item)
            if has_item:
                premises.append(self.problem.condition.get_id_by_predicate_and_item(predicate, item))

            if (not oppose and not has_item) or (oppose and has_item):
                passed = False
                break

        if not passed:
            self.problem.step(self.theorem, 0)
            return

        for equal, item in gpl["algebra_constraints"]:
            oppose = False
            if "~" in equal:
                oppose = True
            eq = get_equation_from_tree(self.problem, item, True, letters)
            solved_eq = False

            result, premise = EqKiller.solve_target(eq, self.problem)
            if result is not None and result == 0:
                solved_eq = True
            premises += premise

            if (not oppose and not solved_eq) or (oppose and solved_eq):
                passed = False
                break

        if not passed:
            self.problem.step(self.theorem, 0)
            return

        for predicate, item in gpl["conclusions"]:
            if predicate == "Equal":  # algebra conclusion
                eq = get_equation_from_tree(self.problem, item, True, letters)
                self.problem.add("Equation", eq, premises, self.theorem)
            else:  # logic conclusion
                item = tuple(letters[i] for i in item)
                self.problem.add(predicate, item, premises, self.theorem)

        EqKiller.solve_equations(self.problem)
        self.problem.step(self.theorem, 0)


class BackwardSearcher:

    def __init__(self, predicate_GDL, theorem_GDL, strategy, max_depth, beam_size, t_info, debug=False):
        """
        Initialize Forward Searcher.
        :param predicate_GDL: predicate GDL.
        :param theorem_GDL: theorem GDL.
        :param strategy: <str>, "dfs", "bfs", "rs", "bs".
        :param max_depth: max search depth.
        :param beam_size: beam search size.
        :param t_info: <dict>, {t_name: (category_id, usage_count)}, user customization.
        :param debug: <bool>, set True when need print process information.
        """
        self.parsed_predicate_GDL = parse_predicate_gdl(predicate_GDL)
        self.parsed_theorem_GDL = parse_theorem_gdl(theorem_GDL, self.parsed_predicate_GDL)
        self.max_depth = max_depth
        self.beam_size = beam_size
        self.strategy = strategy
        self.debug = debug

        self.node_map = None
        self.finder = GoalFinder(self.parsed_theorem_GDL, get_p2t_map_bw(t_info, self.parsed_theorem_GDL))

        self.step_size = None
        self.problem = None
        self.root = None
        self.search_stack = None

        self.id = 0

    def init_search(self, problem_CDL):
        """Init and return a problem by problem_CDL."""
        s_start_time = time.time()
        self.node_map = {}
        self.step_size = 0
        SuperNode.snc = {}

        self.problem = Problem()
        self.problem.load_problem_by_fl(
            self.parsed_predicate_GDL, self.parsed_theorem_GDL, parse_problem_cdl(problem_CDL))  # load problem
        EqKiller.solve_equations(self.problem)
        self.problem.step("init_problem", time.time() - s_start_time)  # save applied theorem and update step

        self.search_stack = []
        self.root = SuperNode(None, self.problem, None, 1, self.node_map, self.finder, self.debug, self.search_stack)
        if self.problem.goal.type == "algebra":
            eq = self.problem.goal.item - self.problem.goal.answer
            self.root.add_nodes([("Equation", eq)])
        else:
            self.root.add_nodes([(self.problem.goal.item, self.problem.goal.answer)])

        self.search_stack.append(self.root)

    def search(self):
        """return seqs, <list> of theorem, solved theorem sequences."""
        pid = self.problem.parsed_problem_CDL["id"]
        debug_print(self.debug, "(pid={}) Start Searching".format(pid))
        if self.strategy == "bfs":
            while self.root.state not in [NodeState.success, NodeState.fail]:
                self.clean_search_stack()
                if len(self.search_stack) == 0:
                    break
                super_node = self.search_stack.pop(0)
                self.step_size += 1
                start_step_count = self.problem.condition.step_count

                timing = time.time()
                debug_print(self.debug, "(pid={},depth={},branch={}/{}) Expanding SuperNode Start".format(
                    pid, super_node.pos[0], super_node.pos[1], SuperNode.snc[super_node.pos[0]]))
                super_node.expand()

                debug_print(self.debug,
                            "(pid={},depth={},branch={}/{}) Expanding SuperNode Done (timing={:.4f})".format(
                                pid, super_node.pos[0], super_node.pos[1], SuperNode.snc[super_node.pos[0]],
                                time.time() - timing))

                timing = time.time()
                debug_print(self.debug, "(pid={},depth={},branch={}/{}) Checking Node Start".format(
                    pid, super_node.pos[0], super_node.pos[1], SuperNode.snc[super_node.pos[0]]))
                self.check_node(start_step_count)
                debug_print(self.debug, "(pid={},depth={},branch={}/{}) Checking Node End (timing={:.4f})".format(
                    pid, super_node.pos[0], super_node.pos[1], SuperNode.snc[super_node.pos[0]],
                    time.time() - timing))
        elif self.strategy == "dfs":
            while self.root.state not in [NodeState.success, NodeState.fail]:
                self.clean_search_stack()
                if len(self.search_stack) == 0:
                    break
                super_node = self.search_stack.pop()
                self.step_size += 1
                start_step_count = self.problem.condition.step_count

                timing = time.time()
                debug_print(self.debug, "(pid={},depth={},branch={}/{}) Expanding SuperNode Start".format(
                    pid, super_node.pos[0], super_node.pos[1], SuperNode.snc[super_node.pos[0]]))
                super_node.expand()

                debug_print(self.debug,
                            "(pid={},depth={},branch={}/{}) Expanding SuperNode Done (timing={:.4f})".format(
                                pid, super_node.pos[0], super_node.pos[1], SuperNode.snc[super_node.pos[0]],
                                time.time() - timing))

                timing = time.time()
                debug_print(self.debug, "(pid={},depth={},branch={}/{}) Checking Node Start".format(
                    pid, super_node.pos[0], super_node.pos[1], SuperNode.snc[super_node.pos[0]]))
                self.check_node(start_step_count)
                debug_print(self.debug, "(pid={},depth={},branch={}/{}) Checking Node End (timing={:.4f})".format(
                    pid, super_node.pos[0], super_node.pos[1], SuperNode.snc[super_node.pos[0]],
                    time.time() - timing))
        elif self.strategy == "rs":
            while self.root.state not in [NodeState.success, NodeState.fail]:
                self.clean_search_stack()
                if len(self.search_stack) == 0:
                    break
                super_node = self.search_stack.pop(random.randint(0, len(self.search_stack) - 1))
                self.step_size += 1
                start_step_count = self.problem.condition.step_count

                timing = time.time()
                debug_print(self.debug, "(pid={},depth={},branch={}/{}) Expanding SuperNode Start".format(
                    pid, super_node.pos[0], super_node.pos[1], SuperNode.snc[super_node.pos[0]]))
                super_node.expand()

                debug_print(self.debug,
                            "(pid={},depth={},branch={}/{}) Expanding SuperNode Done (timing={:.4f})".format(
                                pid, super_node.pos[0], super_node.pos[1], SuperNode.snc[super_node.pos[0]],
                                time.time() - timing))

                timing = time.time()
                debug_print(self.debug, "(pid={},depth={},branch={}/{}) Checking Node Start".format(
                    pid, super_node.pos[0], super_node.pos[1], SuperNode.snc[super_node.pos[0]]))
                self.check_node(start_step_count)
                debug_print(self.debug, "(pid={},depth={},branch={}/{}) Checking Node End (timing={:.4f})".format(
                    pid, super_node.pos[0], super_node.pos[1], SuperNode.snc[super_node.pos[0]],
                    time.time() - timing))
        else:
            while self.root.state not in [NodeState.success, NodeState.fail]:
                self.clean_search_stack()
                if len(self.search_stack) == 0:
                    break
                beam_count = len(self.search_stack)
                if len(self.search_stack) > self.beam_size:  # select branch with beam size
                    search_stack = []
                    for i in random.sample(range(len(self.search_stack)), self.beam_size):
                        search_stack.append(self.search_stack[i])
                    self.search_stack = search_stack
                    beam_count = self.beam_size

                for i in range(beam_count):
                    super_node = self.search_stack.pop(0)
                    if super_node.state != NodeState.to_be_expanded:
                        continue
                    self.step_size += 1
                    start_step_count = self.problem.condition.step_count

                    timing = time.time()
                    debug_print(self.debug, "(pid={},depth={},branch={}/{}) Expanding SuperNode Start".format(
                        pid, super_node.pos[0], super_node.pos[1], SuperNode.snc[super_node.pos[0]]))
                    super_node.expand()

                    debug_print(self.debug,
                                "(pid={},depth={},branch={}/{}) Expanding SuperNode Done (timing={:.4f})".format(
                                    pid, super_node.pos[0], super_node.pos[1], SuperNode.snc[super_node.pos[0]],
                                    time.time() - timing))

                    timing = time.time()
                    debug_print(self.debug, "(pid={},depth={},branch={}/{}) Checking Node Start".format(
                        pid, super_node.pos[0], super_node.pos[1], SuperNode.snc[super_node.pos[0]]))
                    self.check_node(start_step_count)
                    debug_print(self.debug, "(pid={},depth={},branch={}/{}) Checking Node End (timing={:.4f})".format(
                        pid, super_node.pos[0], super_node.pos[1], SuperNode.snc[super_node.pos[0]],
                        time.time() - timing))

                    if self.root.state in [NodeState.success, NodeState.fail]:
                        break

        self.problem.check_goal()
        # self.save_backward_tree()

        if self.problem.goal.solved:
            debug_print(self.debug, "(pid={}) End Searching".format(self.problem.parsed_problem_CDL["id"]))
            _, seqs = get_used_pid_and_theorem(self.problem)
            return True, seqs

        debug_print(self.debug, "(pid={}) End Searching".format(self.problem.parsed_problem_CDL["id"]))
        return False, None

    def clean_search_stack(self):
        for i in range(len(self.search_stack))[::-1]:
            if self.search_stack[i].state == NodeState.to_be_expanded:
                continue
            self.search_stack.pop(i)

    def check_node(self, start_step_count):
        end_step_count = self.problem.condition.step_count
        if start_step_count == end_step_count:
            return

        related_pres = []  # new added predicates
        related_eqs = []  # new added/updated equations
        for step in range(start_step_count, end_step_count):
            for _id in self.problem.condition.ids_of_step[step]:
                if self.problem.condition.items[_id][0] == "Equation":
                    if self.problem.condition.items[_id][1] in related_eqs:
                        continue
                    related_eqs.append(self.problem.condition.items[_id][1])
                    for simp_eq in self.problem.condition.simplified_equation:
                        if simp_eq in related_eqs:
                            continue
                        if _id not in self.problem.condition.simplified_equation[simp_eq]:
                            continue
                        related_eqs.append(simp_eq)
                else:
                    predicate, item = self.problem.condition.items[_id][0:2]
                    if (predicate, item) not in self.node_map or (predicate, item) in related_pres:
                        continue
                    related_pres.append((predicate, item))
        for sym in EqKiller.get_minimum_syms(related_eqs, list(self.problem.condition.simplified_equation)):
            if sym not in self.node_map:
                continue
            related_pres.append(sym)

        for related in related_pres:
            for node in self.node_map[related]:
                if node.state in [NodeState.fail, NodeState.success]:
                    continue
                node.expand(self.search_stack)

        self.check_node(end_step_count)



from formalgeo.parse.basic import parse_geo_predicate, parse_equal_predicate, operation


def parse_theorem_gdl(theorem_GDL, parsed_predicate_GDL):
    """Parse theorem_GDL to logic form."""
    parsed_GDL = {}

    for theorem_name in theorem_GDL:
        name, para, para_len = parse_geo_predicate(theorem_name, True)
        p1 = set(para)

        body = {}
        branch_count = 1
        for branch in theorem_GDL[theorem_name]:
            raw_premise_GDL = [theorem_GDL[theorem_name][branch]["premise"]]
            parsed_premise, paras_list = parse_premise(raw_premise_GDL)
            raw_theorem_GDL = theorem_GDL[theorem_name][branch]["conclusion"]
            parsed_conclusion, paras = parse_conclusion(raw_theorem_GDL)

            p2 = set(paras)  # theorem_GDL format check
            if len(p2 - p1) > 0:
                e_msg = "Theorem GDL definition error in <{}>.".format(theorem_name)
                raise Exception(e_msg)

            for i in range(len(parsed_premise)):
                p2 = set(paras_list[i])
                if len(p2 - p1) > 0 or len(p1 - p2) > 0:
                    e_msg = "Theorem GDL definition error in <{}>.".format(theorem_name)
                    raise Exception(e_msg)
                body[str(branch_count)] = parsed_premise[i]  # add parsed premise
                body[str(branch_count)].update(parsed_conclusion)  # add parsed conclusion
                branch_count += 1

        parsed_GDL[name] = {
            "vars": tuple(para),
            "para_len": tuple(para_len),
            "body": body
        }

    # Auto generate theorem definition for backward solving
    for predicate in parsed_predicate_GDL["Entity"]:
        conclusions = list(parsed_predicate_GDL["Entity"][predicate]["extend"])  # add extend
        for multi in parsed_predicate_GDL["Entity"][predicate]["multi"]:  # add multi
            conclusions.append((predicate, multi))
        if len(conclusions) == 0:
            continue

        name = predicate[0].lower()
        for i in range(1, len(predicate)):
            if predicate[i].isupper():
                name += "_{}".format(predicate[i].lower())
            else:
                name += predicate[i]
        name += "_definition"

        attr_in_conclusions = []
        for conclusion in conclusions:
            if conclusion[0] != "Equal":
                continue
            attr_in_conclusions += find_attrs_in_tree(conclusion)

        parsed_GDL[name] = {
            "vars": parsed_predicate_GDL["Entity"][predicate]["vars"],
            "para_len": parsed_predicate_GDL["Entity"][predicate]["para_len"],
            "body": {
                "1": {
                    "products": ((predicate, parsed_predicate_GDL["Entity"][predicate]["vars"]),),
                    "logic_constraints": (),
                    "algebra_constraints": (),
                    "attr_in_algebra_constraints": (),
                    "conclusions": tuple(conclusions),
                    "attr_in_conclusions": sorted(tuple(set(attr_in_conclusions)))
                }
            }
        }

    for predicate in parsed_predicate_GDL["Relation"]:
        conclusions = list(parsed_predicate_GDL["Relation"][predicate]["extend"])
        for multi in parsed_predicate_GDL["Relation"][predicate]["multi"]:
            conclusions.append((predicate, multi))
        if len(conclusions) == 0:
            continue

        name = predicate[0].lower()
        for i in range(1, len(predicate)):
            if predicate[i].isupper():
                name += "_{}".format(predicate[i].lower())
            else:
                name += predicate[i]
        name += "_definition"

        attr_in_conclusions = []
        for conclusion in conclusions:
            if conclusion[0] != "Equal":
                continue
            attr_in_conclusions += find_attrs_in_tree(conclusion)

        parsed_GDL[name] = {
            "vars": parsed_predicate_GDL["Relation"][predicate]["vars"],
            "para_len": parsed_predicate_GDL["Relation"][predicate]["para_len"],
            "body": {
                "1": {
                    "products": ((predicate, parsed_predicate_GDL["Relation"][predicate]["vars"]),),
                    "logic_constraints": (),
                    "algebra_constraints": (),
                    "attr_in_algebra_constraints": (),
                    "conclusions": tuple(conclusions),
                    "attr_in_conclusions": sorted(tuple(set(attr_in_conclusions)))
                }
            }
        }

    return parsed_GDL


def parse_premise(premise_GDL):
    """
    Parse premise and convert geometric logic statements into disjunctive normal forms.
    'A&(B|C)' ==> 'A&B|A&C' ==> [['A', 'B'], ['A', 'C']]
    >> parse_premise(['Perpendicular(AO,CO)&(Collinear(OBC)|Collinear(OCB))'])
    ([{'products': (('Perpendicular', ('a', 'o', 'c', 'o')), ('Collinear', ('o', 'b', 'c'))),
       'logic_constraints': (),
       'algebra_constraints': (),
       'attr_in_algebra_constraints': ()},
      {'products': (('Perpendicular', ('a', 'o', 'c', 'o')), ('Collinear', ('o', 'c', 'b'))),
       'logic_constraints': (),
       'algebra_constraints': (),
       'attr_in_algebra_constraints': ()}],
     [['a', 'o', 'c', 'o', 'o', 'b', 'c'],
      ['a', 'o', 'c', 'o', 'o', 'c', 'b']])
    """
    update = True
    while update:
        expanded = []
        update = False
        for item in premise_GDL:
            if "|" not in item:
                expanded.append(item)
            else:
                update = True
                left_index = 0
                or_index = item.index("|")
                right_index = -1

                count = 0
                for i in range(1, len(item)):
                    if item[or_index - i] == ")":
                        count -= 1
                    elif item[or_index - i] == "(":
                        count += 1
                    if count == 1:
                        left_index = or_index - i
                        break
                if left_index == 0:
                    head = ""
                else:
                    head = item[0:left_index]

                count = 0
                for i in range(1, len(item)):
                    if item[or_index + i] == "(":
                        count -= 1
                    elif item[or_index + i] == ")":
                        count += 1
                    if count == 1:
                        right_index = or_index + i
                        break
                if right_index == len(item) - 1:
                    tail = ""
                else:
                    tail = item[right_index + 1:len(item)]

                bodies = item[left_index + 1:right_index]
                if "&(" not in bodies:
                    bodies = bodies.split("|")
                else:
                    bodies = bodies.split("|", num=1)
                for body in bodies:
                    expanded.append(head + body + tail)
        premise_GDL = expanded

    paras_list = []
    for i in range(len(premise_GDL)):  # listing
        premise_GDL[i] = premise_GDL[i].split("&")
        attrs = []
        paras = []
        for j in range(len(premise_GDL[i])):
            if "Equal" in premise_GDL[i][j]:
                premise_GDL[i][j], eq_attrs = parse_equal_predicate(premise_GDL[i][j], True)
                attrs += eq_attrs
                for attr, para in eq_attrs:
                    paras += list(para)
            else:
                predicate, para, _ = parse_geo_predicate(premise_GDL[i][j], True)
                premise_GDL[i][j] = (predicate, tuple(para))
                paras += para

        product = []
        logic_constraint = []
        algebra_constraint = []
        existing_paras = set()
        for j in range(len(premise_GDL[i])):
            if premise_GDL[i][j][0][0] == "~":
                continue

            if premise_GDL[i][j][0] == "Equal":
                algebra_constraint.append(premise_GDL[i][j])
            elif len(set(premise_GDL[i][j][1]) - existing_paras) == 0:
                logic_constraint.append(premise_GDL[i][j])
            else:
                existing_paras = existing_paras | set(premise_GDL[i][j][1])
                product.append(premise_GDL[i][j])

        premise_GDL[i] = {
            "products": tuple(product),
            "logic_constraints": tuple(logic_constraint),
            "algebra_constraints": tuple(algebra_constraint),
            "attr_in_algebra_constraints": sorted(tuple([(attr, tuple(para)) for attr, para in set(attrs)]))
        }
        paras_list.append(paras)

    return premise_GDL, paras_list


def parse_conclusion(conclusion_GDL):
    """
    Parse conclusion to logic form, return logic form and para (for format check).
    >> parse_conclusion(['Similar(ABC,ADE)', 'Equal(LengthOfLine(AB),LengthOfLine(CD))'])
    ({'conclusions': (('Similar', ('a', 'b', 'c', 'a', 'd', 'e')),
                     ('Equal', (('LengthOfLine', ('a', 'b')), ('LengthOfLine', ('c', 'd'))))),
      'attr_in_conclusions': (('LengthOfLine', ('c', 'd')),
                             ('LengthOfLine', ('a', 'b')))},
     ['a', 'b', 'c', 'a', 'd', 'e', 'a', 'b', 'c', 'd'])
    """
    paras = []
    attrs = []
    for i in range(len(conclusion_GDL)):
        if "Equal" in conclusion_GDL[i]:
            conclusion_GDL[i], eq_attrs = parse_equal_predicate(conclusion_GDL[i], True)
            attrs += eq_attrs
            for attr, para in eq_attrs:
                paras += list(para)
        else:
            predicate, para, _ = parse_geo_predicate(conclusion_GDL[i], True)
            conclusion_GDL[i] = (predicate, tuple(para))
            paras += para

    parsed_conclusion_GDL = {
        "conclusions": tuple(conclusion_GDL),
        "attr_in_conclusions": sorted(tuple([(attr, tuple(para)) for attr, para in set(attrs)]))
    }
    return parsed_conclusion_GDL, paras


def find_attrs_in_tree(tree):
    """
    Return attrs in parsed equal tree.
    >> find_attrs_in_tree((('Equal', (('Add', (('LengthOfLine', ('o', 'a')), 'x+1')), 'y+2'))))
    [('LengthOfLine', ('o', 'a')]
    """
    attrs = []
    stack = [tree[1][0], tree[1][1]]
    while len(stack) > 0:
        item = stack.pop()
        if not isinstance(item, tuple):
            continue
        if item[0] in operation:
            for item in item[1]:
                stack.append(item)
            continue
        if item not in attrs:
            attrs.append(item)
    return attrs



# Copyright (C) 2022-2023 FormalGeo Development Team
# Author: Xiaokai Zhang
# Contact: formalgeo@gmail.com

"""Download and Management of Datasets and Formal Systems."""

__all__ = [
    "show_available_datasets", "download_dataset", "remove_dataset",
    "DatasetLoader"
]

from formalgeo.data.data import show_available_datasets, download_dataset, remove_dataset
from formalgeo.data.data import DatasetLoader



import sympy
from sympy import sin, cos, tan, sqrt, pi

operation = ["Add", "Sub", "Mul", "Div", "Pow", "Mod", "Sqrt", "Sin", "Cos", "Tan"]


def parse_geo_predicate(s, make_vars=False):
    """
    Parse s to get predicate_name, para, and para structural msg.
    >> parse_geo_predicate('Predicate(ABC)')
    ('Predicate', ['A', 'B', 'C'], [3])
    >> parse_geo_predicate('Predicate(ABC, DE)', True)
    ('Predicate', ['a', 'b', 'c', 'd', 'e'], [3, 2])
    """
    predicate_name, para = s.split("(")
    para = para.replace(")", "")
    if make_vars:
        para = para.lower()

    if "," not in para:
        return predicate_name, list(para), [len(para)]
    para_len = []
    para = para.split(",")
    for item in para:
        para_len.append(len(item))
    return predicate_name, list("".join(para)), para_len


def parse_equal_predicate(s, make_vars=False):
    """
    Parse s to a Tree, return the tree and all attr.
    >> parse_equal_predicate('Equal(LengthOfLine(AB),LengthOfLine(CD))')
    (('Equal', (('LengthOfLine', ('A', 'B')), ('LengthOfLine', ('C', 'D')))),
     [('LengthOfLine', ('A', 'B')), ('LengthOfLine', ('C', 'D'))])
    >> parse_equal_predicate('Equal(LengthOfLine(OA),LengthOfLine(OB))', True)
    (('Equal', (('LengthOfLine', ('o', 'a')), ('LengthOfLine', ('o', 'b')))),
     [('LengthOfLine', ('o', 'a')), ('LengthOfLine', ('o', 'b'))])
    >> parse_equal_predicate('Equal(Add(LengthOfLine(OA),x+1),y+2)', True)
    (('Equal', (('Add', (('LengthOfLine', ('o', 'a')), 'x+1')), 'y+2')),
     [('LengthOfLine', ('o', 'a'))])
    """
    s = s[6:len(s) - 1]
    count = 0
    m = 0
    for i in range(len(s)):
        if s[i] == "(":
            count += 1
        elif s[i] == ")":
            count -= 1

        if count == 0 and s[i] == ",":
            m = i

    if count != 0:
        e_msg = "Sym stack not empty. Miss ')' in {}?.".format(s)
        raise Exception(e_msg)

    left = s[0:m]
    right = s[m + 1:len(s)]
    attrs = []
    if left[0].isupper():
        left, l_attrs = parse_equal_to_tree(left, make_vars)
        attrs += l_attrs
    if right[0].isupper():
        right, l_attrs = parse_equal_to_tree(right, make_vars)
        attrs += l_attrs

    return ("Equal", (left, right)), attrs


def parse_equal_to_tree(s, make_vars=False):
    """
    Parse equal's para to a Tree, return the tree and all attr.
    >> parse_equal_to_tree('LengthOfLine(AB)')
    (('LengthOfLine', ('A', 'B')), [('LengthOfLine', ('A', 'B'))])
    >> parse_equal_to_tree('LengthOfLine(AB)', True)
    (('LengthOfLine', ('a', 'b')), [('LengthOfLine', ('a', 'b'))])
    >> parse_equal_to_tree('Add(LengthOfLine(OA),x+1)', True)
    (('Add', (('LengthOfLine', ('o', 'a')), 'x+1')), [('LengthOfLine', ('o', 'a'))])
    """
    attrs = []
    i = 0
    j = 0
    stack = []
    while j < len(s):
        if s[j] == "(":
            stack.append(s[i:j])
            stack.append(s[j])
            i = j + 1

        elif s[j] == ",":
            if i < j:
                stack.append(s[i: j])
                i = j + 1
            else:
                i = i + 1

        elif s[j] == ")":
            if i < j:
                stack.append(s[i: j])
                i = j + 1
            else:
                i = i + 1

            paras = []
            while stack[-1] != "(":
                paras.append(stack.pop())
            stack.pop()  # pop '('
            predicate = stack.pop()
            if predicate in operation:  # not attribution
                stack.append((predicate, tuple(paras[::-1])))
            else:  # attribution
                paras = tuple("".join(paras[::-1]).lower() if make_vars else "".join(paras[::-1]))
                stack.append((predicate, paras))
                attrs.append((predicate, paras))

        j = j + 1

    return stack.pop(), attrs


def get_equation_from_tree(problem, tree, replaced=False, letters=None):
    """
    Trans expr_tree to symbolic algebraic expression.
    >> get_expr_from_tree(problem, [['LengthOfLine', ['a', 'b']], '2*x-14'], True, {'a': 'Z', 'b': 'X'})
    - 2.0*f_x + ll_zx + 14.0
    >> get_expr_from_tree(problem, [['LengthOfLine', ['Z', 'X']], '2*x-14'])
    - 2.0*f_x + ll_zx + 14.0
    """
    left_expr = get_expr_from_tree(problem, tree[0], replaced, letters)
    if left_expr is None:
        return None
    right_expr = get_expr_from_tree(problem, tree[1], replaced, letters)
    if right_expr is None:
        return None
    return left_expr - right_expr


def get_expr_from_tree(problem, tree, replaced=False, letters=None):
    """
    Recursively trans expr_tree to symbolic algebraic expression.
    :param problem: class <Problem>.
    :param tree: An expression in the form of a list tree.
    :param replaced: Optional. Set True when tree's item is expressed by vars.
    :param letters: Optional. Letters that will replace vars. Dict = {var: letter}.
    >> get_expr_from_tree(problem, ['LengthOfLine', ['T', 'R'[])
    l_tr
    >> get_expr_from_tree(problem, ['Add', [['LengthOfLine', ['Z', 'X']], '2*x-14']])
    2.0*f_x + l_zx - 14.0
    >> get_expr_from_tree(problem, ['Sin', [['MeasureOfAngle', ['a', 'b', 'c']]]],
                          True, {'a': 'X', 'b': 'Y', 'c': 'Z'})
    sin(pi*m_zxy/180)
    """
    if not isinstance(tree, tuple):  # expr
        return parse_expr(problem, tree)
    if tree[0] in problem.parsed_predicate_GDL["Attribution"]:  # attr
        if not replaced:
            return problem.get_sym_of_attr(tree[0], tree[1])
        else:
            replaced_item = [letters[i] for i in tree[1]]
            return problem.get_sym_of_attr(tree[0], tuple(replaced_item))

    if tree[0] in ["Add", "Mul"]:  # operate
        expr_list = []
        for item in tree[1]:
            expr = get_expr_from_tree(problem, item, replaced, letters)
            if expr is None:
                return None
            expr_list.append(expr)
        if tree[0] == "Add":
            result = 0
            for expr in expr_list:
                result += expr
        else:
            result = 1
            for expr in expr_list:
                result *= expr
        return result
    elif tree[0] in ["Sub", "Div", "Pow", "Mod"]:
        expr_left = get_expr_from_tree(problem, tree[1][0], replaced, letters)
        if expr_left is None:
            return None
        expr_right = get_expr_from_tree(problem, tree[1][1], replaced, letters)
        if expr_right is None:
            return None
        if tree[0] == "Sub":
            return expr_left - expr_right
        elif tree[0] == "Div":
            return expr_left / expr_right
        elif tree[0] == "Pow":
            return expr_left ** expr_right
        else:
            return expr_left % expr_right
    elif tree[0] in ["Sin", "Cos", "Tan", "Sqrt"]:
        expr = get_expr_from_tree(problem, tree[1][0], replaced, letters)
        if expr is None:
            return None
        if tree[0] == "Sin":
            return sin(expr * pi / 180)
        elif tree[0] == "Cos":
            return cos(expr * pi / 180)
        elif tree[0] == "Tan":
            return tan(expr * pi / 180)
        else:
            return sqrt(expr)
    else:
        e_msg = "Operator {} not defined, please check your expression.".format(tree[0])
        raise Exception(e_msg)


def parse_expr(problem, expr):
    """Parse expression to symbolic form."""
    expr = sympy.parsing.parse_expr(expr)

    for sym in expr.free_symbols:
        if "_" not in str(sym):
            saved_sym = problem.get_sym_of_attr("Free", str(sym))
            if saved_sym is None:
                return None
            expr = expr.subs(sym, saved_sym)
        else:
            sym_str, para = str(sym).split("_", 1)
            para = tuple(para.upper())
            attr_GDL = problem.parsed_predicate_GDL["Attribution"]
            in_GDL = False
            for attr_name in attr_GDL:
                if attr_GDL[attr_name]["sym"] == sym_str:
                    in_GDL = True
                    saved_sym = problem.get_sym_of_attr(attr_name, para)
                    if saved_sym is None:
                        return None
                    expr = expr.subs(sym, saved_sym)
            if not in_GDL:
                return None

    return expr



import time
import copy
import warnings
from itertools import combinations
from sympy import symbols
from formalgeo.problem.condition import Condition, Goal
from formalgeo.parse import parse_expr, get_equation_from_tree
from formalgeo.tools import rough_equal
from formalgeo.core import EquationKiller as EqKiller


class Problem:
    def __init__(self):
        """Problem conditions, goal, and solving message."""
        self.parsed_predicate_GDL = None
        self.parsed_theorem_GDL = None
        self.parsed_problem_CDL = None
        self.condition = None  # <Condition>, all conditions of current problem.
        self.goal = None  # <Goal>, problem goal.
        self.timing = {}  # <dict>, {step: (theorem, timing)}, such as {0: ('init_problem', 0.00325)}.

    def load_problem_by_fl(self, parsed_predicate_GDL, parsed_theorem_GDL, parsed_problem_CDL):
        """Load problem through problem CDL."""
        self.parsed_predicate_GDL = parsed_predicate_GDL  # gdl
        self.parsed_theorem_GDL = parsed_theorem_GDL  # gdl
        self.parsed_problem_CDL = parsed_problem_CDL  # cdl
        fix_length_predicates = list(self.parsed_predicate_GDL["Preset"]["FixLength"])
        fix_length_predicates += list(self.parsed_predicate_GDL["Entity"])
        fix_length_predicates += list(self.parsed_predicate_GDL["Relation"])
        variable_length_predicates = list(self.parsed_predicate_GDL["Preset"]["VariableLength"])
        self.condition = Condition()
        self.condition.init_by_fl(fix_length_predicates, variable_length_predicates)

        self._construction_init()  # start construction

        # conditions of text_and_image
        for predicate, item in self.parsed_problem_CDL["parsed_cdl"]["text_and_image_cdl"]:
            if predicate == "Equal":
                self.add("Equation", get_equation_from_tree(self, item),
                         (-1,), ("prerequisite", None, None))
            elif predicate == "Equation":
                self.add("Equation", parse_expr(self, item),
                         (-1,), ("prerequisite", None, None))
            else:
                self.add(predicate, tuple(item), (-1,), ("prerequisite", None, None))

        self.goal = Goal()  # set goal
        self.goal.init_by_fl(self, self.parsed_problem_CDL["parsed_cdl"]["goal"])

    def load_problem_by_copy(self, problem):
        """Load problem through copying existing problem."""
        self.parsed_predicate_GDL = problem.parsed_predicate_GDL  # gdl
        self.parsed_theorem_GDL = problem.parsed_theorem_GDL  # gdl
        self.parsed_problem_CDL = problem.parsed_problem_CDL  # cdl
        self.condition = Condition()  # copy all msg of problem
        self.condition.init_by_copy(problem.condition)
        self.timing = copy.deepcopy(problem.timing)
        self.goal = Goal()  # set goal
        self.goal.init_by_copy(problem.goal)

    def _construction_init(self):
        """
        Constructive process.
        1.Collinear expand.
        2.Cocircular expand.
        3.Shape expand. Shape(s1,s2,s3), Shape(s3,s2,s4) ==> Shape(s1,s4).
        4.Angle expand (combination).
        5.Angle expand (collinear).
        6.Angle expand (vertical angle).
        """

        # 1.Collinear expand.
        for predicate, item in self.parsed_problem_CDL["parsed_cdl"]["construction_cdl"]:  # Collinear
            if predicate != "Collinear":
                continue
            if not self.fv_check("Collinear", item):  # FV check
                w_msg = "FV check not passed: [{}, {}]".format(predicate, item)
                warnings.warn(w_msg)
                continue

            added, _id = self.condition.add(predicate, tuple(item), (-1,), ("prerequisite", None, None))
            if not added:
                continue

            self.condition.add(predicate, tuple(item[::-1]), (_id,), ("extended", None, None))
            self.add("Line", (item[0], item[-1]), (_id,), ("extended", None, None))
            for extended_item in combinations(item, 3):  # l=3 is enough
                self.condition.add("Collinear", extended_item, (_id,), ("extended", None, None))
                self.condition.add("Collinear", extended_item[::-1], (_id,), ("extended", None, None))
                self.add("Angle", extended_item, (_id,), ("extended", None, None))
                self.add("Angle", extended_item[::-1], (_id,), ("extended", None, None))

        # 2.Cocircular expand.
        for predicate, item in self.parsed_problem_CDL["parsed_cdl"]["construction_cdl"]:  # Cocircular
            if predicate != "Cocircular":
                continue
            if not self.fv_check("Cocircular", item):  # FV check
                w_msg = "FV check not passed: [{}, {}]".format(predicate, item)
                warnings.warn(w_msg)
                continue

            added, _id = self.condition.add(predicate, tuple(item), (-1,), ("prerequisite", None, None))
            if not added:
                continue

            circle = item[0]
            self.add("Circle", (circle,), (_id,), ("extended", None, None))
            if len(item) == 1:
                continue

            item = item[1:]
            for com in range(1, len(item) + 1):  # extend cocircular
                for extended_item in combinations(item, com):
                    if com == 2:
                        self.condition.add("Arc", (circle, extended_item[0], extended_item[-1]),
                                           (_id,), ("extended", None, None))
                        self.condition.add("Arc", (circle, extended_item[-1], extended_item[0]),
                                           (_id,), ("extended", None, None))
                    cocircular = list(extended_item)
                    l = len(cocircular)
                    for bias in range(l):
                        extended_item = tuple([circle] + [cocircular[(i + bias) % l] for i in range(l)])
                        self.condition.add("Cocircular", extended_item, (_id,), ("extended", None, None))

        # 3.Shape expand.
        jigsaw_unit = {}  # shape's jigsaw
        shape_unit = []  # mini shape unit
        for predicate, item in self.parsed_problem_CDL["parsed_cdl"]["construction_cdl"]:  # Shape
            if predicate != "Shape":
                continue
            if not self.fv_check("Shape", item):  # FV check
                w_msg = "FV check not passed: [{}, {}]".format(predicate, item)
                warnings.warn(w_msg)
                continue

            if len(item) == 1:  # point or line
                if len(item[0]) == 1:
                    self.add("Point", tuple(item[0]), (-1,), ("prerequisite", None, None))
                else:
                    self.add("Line", tuple(item[0]), (-1,), ("prerequisite", None, None))
                continue
            elif len(item) == 2 and len(item[0]) == 2 and len(item[1]) == 2:  # angle
                self.add("Angle", tuple(item[0] + item[1][1]), (-1,), ("prerequisite", None, None))
                continue

            added, all_forms = self._add_shape(tuple(item), (-1,), ("prerequisite", None, None))  # shape
            if not added:
                continue

            for shape in all_forms:
                jigsaw_unit[shape] = all_forms
                shape_unit.append(shape)

        shape_comb = shape_unit
        jigsaw_comb = jigsaw_unit
        while len(shape_comb):
            shape_comb_new = []
            jigsaw_comb_new = {}
            for unit in shape_unit:
                for comb in shape_comb:

                    if len(unit[-1]) != len(comb[0]):  # has same sides?
                        continue
                    elif len(unit[-1]) == 3:  # is arc and same?
                        if unit[-1] != comb[0]:
                            continue
                    else:
                        if unit[-1] != comb[0][::-1]:  # is line and same?
                            continue

                    if unit in jigsaw_comb[comb]:  # comb is combined from unit
                        continue

                    same_length = 1  # number of same sides
                    mini_length = len(unit) if len(unit) < len(comb) else len(comb)  # mini length
                    while same_length < mini_length:
                        if len(unit[- same_length - 1]) != len(comb[same_length]):  # all arcs or all lines
                            break
                        elif len(unit[- same_length - 1]) == 3:  # arc
                            if unit[- same_length - 1] != comb[same_length]:
                                break
                        else:  # line
                            if unit[- same_length - 1] != comb[same_length][::-1]:
                                break

                        same_length += 1

                    new_shape = list(unit[0:len(unit) - same_length])  # diff sides in polygon1
                    new_shape += list(comb[same_length:len(comb)])  # diff sides in polygon2

                    if not len(new_shape) == len(set(new_shape)):  # ensure no ring
                        continue

                    new_shape = tuple(new_shape)
                    if self.condition.has("Shape", new_shape):
                        continue

                    all_sides = ""
                    for item in new_shape:  # remove circle center point
                        if len(item) == 3:
                            item = item[1:]
                        all_sides += item
                    checked = True
                    for point in all_sides:
                        if all_sides.count(point) > 2:
                            checked = False
                            break
                    if not checked:  # ensure no holes
                        continue

                    premise = (self.condition.get_id_by_predicate_and_item("Shape", unit),
                               self.condition.get_id_by_predicate_and_item("Shape", comb))

                    added, all_forms = self._add_shape(new_shape, premise, ("extended", None, None))  # add shape
                    if not added:  # ensure added
                        continue

                    new_shape_jigsaw = jigsaw_unit[unit] | jigsaw_comb[comb]
                    for shape in all_forms:
                        jigsaw_comb_new[shape] = new_shape_jigsaw
                        shape_comb_new.append(shape)

            shape_comb = shape_comb_new
            jigsaw_comb = jigsaw_comb_new

        # 4.Angle expand (combination).
        angle_unit = self.condition.get_items_by_predicate("Angle")
        jigsaw_unit = {}
        for angle in angle_unit:
            jigsaw_unit[angle] = {angle}

        angle_comb = angle_unit  # combination angle
        jigsaw_comb = jigsaw_unit  # angle's jigsaw
        while len(angle_comb):
            angle_comb_new = []
            jigsaw_comb_new = {}
            for unit in angle_unit:
                for comb in angle_comb:

                    if unit in jigsaw_comb[comb]:  # comb is combined from unit
                        continue

                    if not (unit[1] == comb[1] and unit[2] == comb[0] and unit[0] != comb[2]):  # ensure adjacent
                        continue

                    angles = self.condition.get_items_by_predicate("Angle")
                    if (unit[0], unit[1], comb[2]) in angles or \
                            (unit[0], comb[2], unit[1]) in angles or \
                            (comb[2], unit[0], unit[1]) in angles:
                        continue

                    new_angle = (unit[0], unit[1], comb[2])

                    if not len(new_angle) == len(set(new_angle)):  # ensure same points
                        continue

                    premise = (self.condition.get_id_by_predicate_and_item("Angle", unit),
                               self.condition.get_id_by_predicate_and_item("Angle", comb))
                    added, _ = self.condition.add("Angle", new_angle, premise,
                                                  ("extended", None, None))  # need to expand line
                    if not added:
                        continue

                    new_angle_jigsaw = jigsaw_unit[unit] | jigsaw_comb[comb]
                    jigsaw_comb_new[new_angle] = new_angle_jigsaw
                    angle_comb_new.append(new_angle)

            angle_comb = angle_comb_new
            jigsaw_comb = jigsaw_comb_new

        # 5.Angle collinear expand.
        for angle in self.condition.get_items_by_predicate("Angle"):
            a, v, b = angle
            a_collinear = None
            b_collinear = None
            for predicate, item in self.parsed_problem_CDL["parsed_cdl"]["construction_cdl"]:
                if predicate == "Collinear" and v in item:
                    if a in item:
                        a_collinear = item
                    if b in item:
                        b_collinear = item

            a_points = []  # Points collinear with a and on the same side with a
            b_points = []
            if a_collinear is not None:
                if a_collinear.index(v) < a_collinear.index(a):  # .....V...P..
                    i = a_collinear.index(v) + 1
                    while i < len(a_collinear):
                        a_points.append(a_collinear[i])
                        i += 1
                else:  # ...P.....V...
                    i = 0
                    while i < a_collinear.index(v):
                        a_points.append(a_collinear[i])
                        i += 1
            else:
                a_points.append(a)

            if b_collinear is not None:
                if b_collinear.index(v) < b_collinear.index(b):  # .....V...P..
                    i = b_collinear.index(v) + 1
                    while i < len(b_collinear):
                        b_points.append(b_collinear[i])
                        i += 1
                else:  # ...P.....V...
                    i = 0
                    while i < b_collinear.index(v):
                        b_points.append(b_collinear[i])
                        i += 1
            else:
                b_points.append(b)

            for a_point in a_points:
                for b_point in b_points:
                    premise = (self.condition.get_id_by_predicate_and_item("Angle", angle),)
                    self.add("Angle", (a_point, v, b_point), premise, ("extended", None, None))

        # 6.Angle expand (vertical angle).
        for angle in self.condition.get_items_by_predicate("Angle"):
            premise = (self.condition.get_id_by_predicate_and_item("Angle", angle),)
            self.add("Angle", (angle[2], angle[1], angle[0]), premise, ("extended", None, None))

    def _add_shape(self, shape, premise, theorem):
        """pass"""
        added, _id = self.condition.add("Shape", shape, premise, theorem)
        if not added:
            return False, None

        all_forms = [shape]
        l = len(shape)
        for bias in range(1, l):  # all forms
            new_item = tuple([shape[(i + bias) % l] for i in range(l)])
            self.condition.add("Shape", new_item, (_id,), ("extended", None, None))
            all_forms.append(new_item)

        shape = list(shape)
        _, col = self.condition.get_ids_and_items_by_predicate_and_variable("Collinear", ["a", "b", "c"])
        i = 0
        has_arc = False
        while i < len(shape):
            if len(shape[i]) == 2:
                self.add("Line", (shape[i][0], shape[i][1]),
                         (-1,), ("prerequisite", None, None))
            else:
                has_arc = True
                i += 1
                continue

            j = (i + 1) % len(shape)
            if len(shape[j]) == 2:
                self.add("Angle", (shape[i][0], shape[i][1], shape[j][1]),
                         (-1,), ("prerequisite", None, None))  # extend angle
                if (shape[i][0], shape[i][1], shape[j][1]) in col:
                    shape[i] = shape[i][0] + shape[j][1]
                    shape.pop(j)
                    continue
            i += 1

        if not has_arc and len(shape) > 2:  # extend polygon
            valid = True
            i = 0
            l = len(shape)
            while valid and i < l:
                if shape[i][1] != shape[(i + 1) % l][0]:
                    valid = False
                i += 1
            if valid:
                self.add("Polygon", tuple([item[0] for item in shape]),
                         (-1,), ("prerequisite", None, None))

        return True, set(all_forms)

    def _align_angle_sym(self, angle):
        """
        Make the symbols of angles the same.
        Measure(Angle(ABC)), Measure(Angle(ABD))  ==>  m_abc,  if Collinear(BCD)
        """
        sym = symbols("ma_" + "".join(angle).lower(), positive=True)  # init sym
        self.condition.value_of_sym[sym] = None  # init symbol's value

        a, v, b = angle
        a_collinear = None
        b_collinear = None
        for predicate, item in self.parsed_problem_CDL["parsed_cdl"]["construction_cdl"]:
            if predicate == "Collinear" and v in item:
                if a in item:
                    a_collinear = item
                if b in item:
                    b_collinear = item

        a_points = []  # Points collinear with a and on the same side with a
        b_points = []
        if a_collinear is not None:
            if a_collinear.index(v) < a_collinear.index(a):  # .....V...P..
                i = a_collinear.index(v) + 1
                while i < len(a_collinear):
                    a_points.append(a_collinear[i])
                    i += 1
            else:  # ...P.....V...
                i = 0
                while i < a_collinear.index(v):
                    a_points.append(a_collinear[i])
                    i += 1
        else:
            a_points.append(a)

        if b_collinear is not None:
            if b_collinear.index(v) < b_collinear.index(b):  # .....V...P..
                i = b_collinear.index(v) + 1
                while i < len(b_collinear):
                    b_points.append(b_collinear[i])
                    i += 1
            else:  # ...P.....V...
                i = 0
                while i < b_collinear.index(v):
                    b_points.append(b_collinear[i])
                    i += 1
        else:
            b_points.append(b)

        same_angles = []  # Same angle get by collinear
        for a_point in a_points:
            for b_point in b_points:
                same_angles.append((a_point, v, b_point))

        for same_angle in same_angles:
            self.condition.sym_of_attr[("MeasureOfAngle", same_angle)] = sym
        self.condition.attr_of_sym[sym] = ("MeasureOfAngle", tuple(same_angles))

        return sym

    def add(self, predicate, item, premise, theorem, skip_check=False):
        """
        Add item to condition of specific predicate category.
        Also consider condition expansion and equation construction.
        :param predicate: Construction, Entity, Relation or Equation.
        :param item: <tuple> or equation.
        :param premise: tuple of <int>, premise of item.
        :param theorem: <tuple>, (t_name, t_branch, t_para).
        :param skip_check: <bool>, set to True when you are confident that the format of item must be legal.
        :return: True or False.
        """
        if not skip_check and not self.check(predicate, item, premise, theorem):
            return False

        added, _id = self.condition.add(predicate, item, premise, theorem)
        if added:
            if predicate == "Equation":  # preset Equation
                return True

            if predicate in self.parsed_predicate_GDL["Preset"]["BasicEntity"]:  # preset BasicEntity
                if predicate == "Line":
                    self.condition.add("Line", item[::-1], (_id,), ("extended", None, None))
                    self.condition.add("Point", (item[0],), (_id,), ("extended", None, None))
                    self.condition.add("Point", (item[1],), (_id,), ("extended", None, None))
                elif predicate == "Arc":
                    self.condition.add("Point", (item[1],), (_id,), ("extended", None, None))
                    self.condition.add("Point", (item[2],), (_id,), ("extended", None, None))
                elif predicate == "Angle":
                    self.add("Line", (item[0], item[1]),
                             (_id,), ("extended", None, None), skip_check=True)
                    self.add("Line", (item[1], item[2]),
                             (_id,), ("extended", None, None), skip_check=True)
                elif predicate == "Polygon":
                    l = len(item)
                    for bias in range(1, l):  # all forms
                        new_item = tuple([item[(i + bias) % l] for i in range(l)])
                        self.condition.add("Polygon", new_item, (_id,), ("extended", None, None))
                return True  # Point and Circle no need to extend

            if predicate in self.parsed_predicate_GDL["Entity"]:  # user defined Entity
                item_GDL = self.parsed_predicate_GDL["Entity"][predicate]
            else:  # user defined Relation
                item_GDL = self.parsed_predicate_GDL["Relation"][predicate]

            predicate_vars = item_GDL["vars"]
            letters = {}  # used for vars-letters replacement
            for i in range(len(predicate_vars)):
                letters[predicate_vars[i]] = item[i]

            for para_list in item_GDL["multi"]:  # multi
                self.condition.add(predicate, tuple(letters[i] for i in para_list), (_id,), ("extended", None, None))

            for extended_predicate, para in item_GDL["extend"]:  # extended
                if extended_predicate == "Equal":
                    self.add("Equation", get_equation_from_tree(self, para, True, letters),
                             (_id,), ("extended", None, None))
                else:
                    self.add(extended_predicate, tuple(letters[i] for i in para),
                             (_id,), ("extended", None, None))

            return True

        return False

    def check(self, predicate, item, premise=None, theorem=None):
        """
        EE check and FV check.
        :param predicate: Construction, Entity, Relation or Equation.
        :param item: <tuple> or equation.
        :param premise: tuple of <int>, premise of item.
        :param theorem: <str>, theorem of item.
        :return: True or False.
        """
        if predicate not in self.condition.items_group:  # predicate must be defined
            e_msg = "Predicate '{}' not defined in current predicate GDL.".format(predicate)
            raise Exception(e_msg)
        if not self.ee_check(predicate, item):  # ee check
            w_msg = "EE check not passed: [{}, {}, {}, {}]".format(predicate, item, premise, theorem)
            warnings.warn(w_msg)
            return False
        if not self.fv_check(predicate, item):  # fv check
            w_msg = "FV check not passed: [{}, {}, {}, {}]".format(predicate, item, premise, theorem)
            warnings.warn(w_msg)
            return False

        return True

    def ee_check(self, predicate, item):
        """Entity Existence check."""

        if predicate == "Equation" or predicate in self.parsed_predicate_GDL["Preset"]["BasicEntity"] \
                or predicate in self.parsed_predicate_GDL["Preset"]["Construction"]:
            return True
        elif predicate in self.parsed_predicate_GDL["Entity"]:
            item_GDL = self.parsed_predicate_GDL["Entity"][predicate]
        elif predicate in self.parsed_predicate_GDL["Relation"]:
            item_GDL = self.parsed_predicate_GDL["Relation"][predicate]
        elif predicate == "Free":
            return True
        else:
            item_GDL = self.parsed_predicate_GDL["Attribution"][predicate]

        letters = {}  # used for vars-letters replacement
        for i in range(len(item_GDL["vars"])):
            letters[item_GDL["vars"][i]] = item[i]

        for name, para in item_GDL["ee_check"]:
            if tuple(letters[i] for i in para) not in self.condition.get_items_by_predicate(name):
                return False
        return True

    def fv_check(self, predicate, item):
        """Format Validity check."""

        if predicate == "Equation":
            if item is None or item == 0:
                return False
            return True
        elif predicate in self.parsed_predicate_GDL["Preset"]["Construction"]:
            if predicate == "Shape":
                if len(item) != len(set(item)):  # default check 1: mutex points
                    return False
                if len(item) == 1:
                    if len(item[0]) not in [1, 2]:
                        return False
                    return True
                for shape in item:
                    if not 2 <= len(shape) <= 3 or len(shape) != len(set(shape)):
                        return False
                return True
            else:
                return len(item) == len(set(item))  # default check 1: mutex points
        elif predicate in self.parsed_predicate_GDL["Preset"]["BasicEntity"]:
            if len(item) != len(set(item)):  # default check 1: mutex points
                return False
            if predicate == "Point" and len(item) != 1:
                return False
            elif predicate == "Line" and len(item) != 2:
                return False
            elif predicate == "Arc" and len(item) != 3:
                return False
            elif predicate == "Angle" and len(item) != 3:
                return False
            elif predicate == "Polygon" and len(item) < 3:
                return False
            elif predicate == "Circle" and len(item) != 1:
                return False
            return True
        elif predicate in self.parsed_predicate_GDL["Entity"]:
            if len(item) != len(set(item)):  # default check 1: mutex points
                return False
            item_GDL = self.parsed_predicate_GDL["Entity"][predicate]
        elif predicate in self.parsed_predicate_GDL["Relation"]:
            item_GDL = self.parsed_predicate_GDL["Relation"][predicate]
        elif predicate == "Free":
            return True
        else:
            item_GDL = self.parsed_predicate_GDL["Attribution"][predicate]

        if len(item) != len(item_GDL["vars"]):  # default check 2: correct para len
            return False

        if "fv_check" in item_GDL:  # fv check, more stringent than default check 3
            checked = []
            result = []
            for i in item:
                if i not in checked:
                    checked.append(i)
                result.append(str(checked.index(i)))
            if "".join(result) in item_GDL["fv_check"]:
                return True
            return False

        if len(item_GDL["ee_check"]) > 1:  # default check 3: para of the same type need to be different
            predicate_to_vars = {}
            for predicate, p_var in item_GDL["ee_check"]:
                if predicate not in self.parsed_predicate_GDL["Preset"]["Construction"]:  # check only BasicEntity
                    if predicate not in predicate_to_vars:
                        predicate_to_vars[predicate] = [p_var]
                    else:
                        predicate_to_vars[predicate].append(p_var)

            letters = {}  # used for vars-letters replacement
            for i in range(len(item_GDL["vars"])):
                letters[item_GDL["vars"][i]] = item[i]

            for predicate in predicate_to_vars:
                if len(predicate_to_vars[predicate]) == 1:
                    continue

                mutex_sets = []  # mutex_item
                for p_var in predicate_to_vars[predicate]:
                    mutex_sets.append([letters[i] for i in p_var])

                mutex_sets_multi = []  # mutex_item multi representation
                for mutex_item in mutex_sets:
                    if predicate == "Line":
                        mutex_sets_multi.append(tuple(mutex_item))
                        mutex_sets_multi.append(tuple(mutex_item[::-1]))
                    elif predicate == "Polygon":
                        l = len(mutex_item)
                        for bias in range(0, l):
                            mutex_sets_multi.append(tuple([mutex_item[(i + bias) % l] for i in range(l)]))
                    else:  # Point Arc Angle Circle
                        mutex_sets_multi.append(tuple(mutex_item))

                if len(mutex_sets_multi) != len(set(mutex_sets_multi)):
                    return False

        return True

    def get_sym_of_attr(self, attr, item):
        """
        Get symbolic representation of item's attribution.
        :param attr: attr's name, such as LengthOfLine
        :param item: tuple, such as ('A', 'B')
        :return: sym
        """

        if attr != "Free" and attr not in self.parsed_predicate_GDL["Attribution"]:  # attr must define
            e_msg = "Attribution '{}' not defined in current predicate GDL.".format(attr)
            raise Exception(e_msg)
        if not self.ee_check(attr, item):  # ee check
            msg = "EE check not passed: [{}, {}]".format(attr, item)
            warnings.warn(msg)
            return None
        if not self.fv_check(attr, item):  # fv check
            msg = "FV check not passed: [{}, {}]".format(attr, item)
            warnings.warn(msg)
            return None

        if (attr, item) in self.condition.sym_of_attr:  # already has sym
            return self.condition.sym_of_attr[(attr, item)]

        if attr == "Free":
            sym = symbols("".join(item))
            self.condition.sym_of_attr[(attr, item)] = sym  # add sym
            self.condition.value_of_sym[sym] = None  # init symbol's value
            self.condition.attr_of_sym[sym] = (attr, (item,))  # add attr
            return sym

        if attr == "MeasureOfAngle":  # align angle's sym
            return self._align_angle_sym(item)

        attr_GDL = self.parsed_predicate_GDL["Attribution"][attr]
        if (attr, item) not in self.condition.sym_of_attr:  # No symbolic representation, initialize one.
            sym = symbols(attr_GDL["sym"] + "_" + "".join(item).lower(), positive=True)
            self.condition.sym_of_attr[(attr, item)] = sym  # add sym
            self.condition.value_of_sym[sym] = None  # init symbol's value

            extend_items = [item]

            letters = {}  # used for vars-letters replacement
            for i in range(len(attr_GDL["vars"])):
                letters[attr_GDL["vars"][i]] = item[i]

            for multi in attr_GDL["multi"]:
                extended_item = [letters[i] for i in multi]  # extend item
                self.condition.sym_of_attr[(attr, tuple(extended_item))] = sym  # multi representation
                extend_items.append(tuple(extended_item))

            self.condition.attr_of_sym[sym] = (attr, tuple(extend_items))  # add attr
            return sym

    def set_value_of_sym(self, sym, value, premise):
        """
        Set value of sym.
        Add equation to record the premise and theorem of solving the symbol's value at the same time.
        :param sym: <symbol>
        :param value: <float>
        :param premise: tuple of <int>, premise of getting value.
        """

        if self.condition.value_of_sym[sym] is None:
            self.condition.value_of_sym[sym] = value
            added, _id = self.condition.add("Equation", sym - value, premise, ("solve_eq", None, None))
            return added
        return False

    def step(self, item, timing):
        """
        Execute when theorem successful applied. Save theorem and update step.
        :param item: <str>, theorem, 'init_problem' and 'check_goal'.
        :param timing: <float>.
        """
        self.timing[self.condition.step_count] = (item, timing)
        self.condition.step()

    def check_goal(self):
        """Check whether the solution is completed."""
        s_start_time = time.time()  # timing
        if self.goal.type == "algebra":  # algebra relation
            try:
                result, premise = EqKiller.solve_target(self.goal.item, self)
            except BaseException as e:
                msg = "Exception occur when solve target: {}".format(repr(e))
                warnings.warn(msg)
            else:
                if result is not None:
                    if rough_equal(result, self.goal.answer):
                        self.goal.solved = True
                    self.goal.solved_answer = result

                    eq = self.goal.item - result
                    if eq in self.condition.get_items_by_predicate("Equation"):
                        self.goal.premise = self.condition.get_premise_by_predicate_and_item("Equation", eq)
                        self.goal.theorem = self.condition.get_theorem_by_predicate_and_item("Equation", eq)
                    else:
                        self.goal.premise = tuple(premise)
                        self.goal.theorem = ("solve_eq", None, None)
        elif self.goal.type == "logic":  # logic relation
            if self.goal.answer in self.condition.get_items_by_predicate(self.goal.item):
                self.goal.solved = True
                self.goal.solved_answer = self.goal.answer
                self.goal.premise = self.condition.get_premise_by_predicate_and_item(self.goal.item, self.goal.answer)
                self.goal.theorem = self.condition.get_theorem_by_predicate_and_item(self.goal.item, self.goal.answer)

        self.step("check_goal", time.time() - s_start_time)



from formalgeo.solver import Interactor, ForwardSearcher, BackwardSearcher
from formalgeo.tools import load_json, save_json
from formalgeo.tools import show_solution, draw_solution_tree_and_theorem_dag
from formalgeo.parse import parse_theorem_seqs, inverse_parse_one_theorem
from formalgeo.tools import get_solution_step, get_solution_hypertree, get_theorem_dag, get_used_pid_and_theorem
from formalgeo.data import DatasetLoader
import random
import warnings

if __name__ == '__main__':

    test_datasets_path = "F:/Datasets/released"
    # show_available_datasets(test_datasets_path)

    dl = DatasetLoader("formalgeo-imo_v1", test_datasets_path)
    solver = Interactor(dl.predicate_GDL, dl.theorem_GDL)

    problem_CDL = dl.get_problem(7)
    solver.load_problem(problem_CDL)
    for t_name, t_branch, t_para in parse_theorem_seqs(problem_CDL["theorem_seqs"]):
        solver.apply_theorem(t_name, t_branch, t_para)
    solver.problem.check_goal()  # check goal after applied theorem seqs

    save_json(get_solution_hypertree(solver.problem), "7_hypertree.json")
    save_json(get_theorem_dag(solver.problem), "7_dag.json")
    draw_solution_tree_and_theorem_dag(solver.problem, "./")



from formalgeo.parse.basic import parse_geo_predicate, parse_equal_predicate


def parse_predicate_gdl(predicate_GDL):
    """parse predicate_GDL to logic form."""
    parsed_GDL = {  # preset Construction
        "Preset": predicate_GDL["Preset"],
        "Entity": {},
        "Relation": {},
        "Attribution": {}
    }
    entities = predicate_GDL["Entity"]  # parse entity
    for item in entities:
        name, para, para_len = parse_geo_predicate(item, True)
        parsed_GDL["Entity"][name] = {
            "vars": tuple(para),
            "para_len": tuple(para_len),
            "ee_check": tuple(parse_ee_check(entities[item]["ee_check"])),
            "multi": tuple(parse_multi(entities[item]["multi"])),
            "extend": tuple(parse_extend(entities[item]["extend"]))
        }

    relations = predicate_GDL["Relation"]  # parse relation
    for item in relations:
        name, para, para_len = parse_geo_predicate(item, True)
        if "fv_check" in relations[item]:
            parsed_GDL["Relation"][name] = {
                "vars": tuple(para),
                "para_len": tuple(para_len),
                "ee_check": tuple(parse_ee_check(relations[item]["ee_check"])),
                "fv_check": tuple(parse_fv_check(relations[item]["fv_check"])),
                "multi": tuple(parse_multi(relations[item]["multi"])),
                "extend": tuple(parse_extend(relations[item]["extend"]))
            }
        else:
            parsed_GDL["Relation"][name] = {
                "vars": tuple(para),
                "para_len": tuple(para_len),
                "ee_check": tuple(parse_ee_check(relations[item]["ee_check"])),
                "multi": tuple(parse_multi(relations[item]["multi"])),
                "extend": tuple(parse_extend(relations[item]["extend"]))
            }

    attributions = predicate_GDL["Attribution"]  # parse attribution
    for item in attributions:
        name, para, para_len = parse_geo_predicate(item, True)
        if "fv_check" in attributions[item]:
            parsed_GDL["Attribution"][name] = {
                "vars": tuple(para),
                "para_len": tuple(para_len),
                "ee_check": tuple(parse_ee_check(attributions[item]["ee_check"])),
                "fv_check": tuple(parse_fv_check(attributions[item]["fv_check"])),
                "sym": attributions[item]["sym"],
                "multi": tuple(parse_multi(attributions[item]["multi"]))
            }
        else:
            parsed_GDL["Attribution"][name] = {
                "vars": tuple(para),
                "para_len": tuple(para_len),
                "ee_check": tuple(parse_ee_check(attributions[item]["ee_check"])),
                "sym": attributions[item]["sym"],
                "multi": tuple(parse_multi(attributions[item]["multi"]))
            }

    return parsed_GDL


def parse_ee_check(ee_check):
    """
    parse ee_check to logic form.
    >> parse_ee_check(['Triangle(ABC)'])
    [('Triangle', ('a', 'b', 'c'))]
    >> parse_ee_check(['Line(AO)', 'Line(CO)'])
    [('Line', ('a', 'o')), ('Line', ('c', 'o'))]
    """
    results = []
    for item in ee_check:
        name, item_para, _ = parse_geo_predicate(item, True)
        results.append((name, tuple(item_para)))
    return results


def parse_fv_check(fv_check):
    """
    parse fv_check to logic form.
    >> parse_fv_check(['O,AB,CD'])
    ['01234']
    >> parse_fv_check(['AD,ABC', 'AB,ABC', 'AC,ABC'])
    ['01023', '01012', '01021']
    """
    results = []
    for item in fv_check:
        checked = []
        result = []
        for i in item.replace(",", ""):
            if i not in checked:
                checked.append(i)
            result.append(str(checked.index(i)))
        results.append("".join(result))
    return results


def parse_multi(multi):
    """
    parse multi to logic form.
    >> _parse_multi(['BCA', 'CAB'])
    [('b', 'c', 'a'), ('c', 'a', 'b')]
    >> _parse_multi(['M,BA'])
    [('m', 'b', 'a')]
    """
    return [tuple(parsed_multi.replace(",", "").lower()) for parsed_multi in multi]


def parse_extend(extend):
    """
    parse extend to logic form.
    >> parse_extend(['Equal(MeasureOfAngle(AOC),90)'])
    [('Equal', (('MeasureOfAngle', ('a', 'o', 'c')), '90'))]
    >> parse_extend(['Perpendicular(AB,CB)', 'IsAltitude(AB,ABC)'])
    [('Perpendicular', ('a', 'b', 'c', 'b')), ('IsAltitude', ('a', 'b', 'a', 'b', 'c'))]
    """
    results = []
    for extend_item in extend:
        if extend_item.startswith("Equal"):
            parsed_equal, _ = parse_equal_predicate(extend_item, True)
            results.append(parsed_equal)
        else:
            extend_name, extend_para, _ = parse_geo_predicate(extend_item, True)
            results.append((extend_name, tuple(extend_para)))
    return results



from formalgeo.parse.basic import parse_geo_predicate, parse_equal_predicate, parse_equal_to_tree


def parse_problem_cdl(problem_CDL):
    """parse problem_CDL to logic form."""
    parsed_CDL = {
        "id": problem_CDL["problem_id"],
        "cdl": {
            "construction_cdl": problem_CDL["construction_cdl"],
            "text_cdl": problem_CDL["text_cdl"],
            "image_cdl": problem_CDL["image_cdl"],
            "goal_cdl": problem_CDL["goal_cdl"]
        },
        "parsed_cdl": {
            "construction_cdl": [],
            "text_and_image_cdl": [],
            "goal": {},
        }
    }
    for fl in problem_CDL["construction_cdl"]:
        predicate, para = fl.split("(")
        para = para.replace(")", "")
        if predicate == "Shape":
            parsed_CDL["parsed_cdl"]["construction_cdl"].append([predicate, para.split(",")])
        elif predicate == "Collinear":
            parsed_CDL["parsed_cdl"]["construction_cdl"].append([predicate, list(para)])
        elif predicate == "Cocircular":
            parsed_CDL["parsed_cdl"]["construction_cdl"].append([predicate, list(para.replace(",", ""))])
        else:
            e_msg = "The predicate <{}> should not appear in construction_cdl.".format(predicate)
            raise Exception(e_msg)

    for fl in problem_CDL["text_cdl"] + problem_CDL["image_cdl"]:
        if fl.startswith("Equal"):
            parsed_equal, _ = parse_equal_predicate(fl)
            parsed_CDL["parsed_cdl"]["text_and_image_cdl"].append(parsed_equal)
        elif fl.startswith("Equation"):
            fl = fl.replace("Equation(", "")
            fl = fl[0:len(fl) - 1]
            parsed_CDL["parsed_cdl"]["text_and_image_cdl"].append(("Equation", fl))
        else:
            predicate, para, _ = parse_geo_predicate(fl)
            parsed_CDL["parsed_cdl"]["text_and_image_cdl"].append([predicate, para])

    if problem_CDL["goal_cdl"].startswith("Value"):
        parsed_CDL["parsed_cdl"]["goal"]["type"] = "value"
        parsed_goal = problem_CDL["goal_cdl"][6:len(problem_CDL["goal_cdl"]) - 1]
        if parsed_goal[0].isupper():
            parsed_goal, _ = parse_equal_to_tree(parsed_goal)
            parsed_CDL["parsed_cdl"]["goal"]["item"] = ("Value", parsed_goal)
        else:
            parsed_CDL["parsed_cdl"]["goal"]["item"] = ("Value", parsed_goal)
        parsed_CDL["parsed_cdl"]["goal"]["answer"] = problem_CDL["problem_answer"]
    elif problem_CDL["goal_cdl"].startswith("Equal"):
        parsed_CDL["parsed_cdl"]["goal"]["type"] = "equal"
        parsed_goal, _ = parse_equal_predicate(problem_CDL["goal_cdl"])
        parsed_CDL["parsed_cdl"]["goal"]["item"] = parsed_goal
        parsed_CDL["parsed_cdl"]["goal"]["answer"] = "0"
    elif problem_CDL["goal_cdl"].startswith("Relation"):
        parsed_CDL["parsed_cdl"]["goal"]["type"] = "logic"
        predicate, para, _ = parse_geo_predicate(
            problem_CDL["goal_cdl"].split("(", 1)[1])
        parsed_CDL["parsed_cdl"]["goal"]["item"] = predicate
        parsed_CDL["parsed_cdl"]["goal"]["answer"] = para

    return parsed_CDL


def parse_theorem_seqs(theorem_seqs):
    """Parse theorem_seqs to logic form."""
    results = []

    for theorem in theorem_seqs:
        results.append(parse_one_theorem(theorem))

    return results


def parse_one_theorem(theorem):
    """
    Parse one theorem to logic form.
    >> parse_one_theorem('congruent_triangle_property_angle_equal(1,RST,XYZ)')
    ('congruent_triangle_property_angle_equal', 1, ('R', 'S', 'T', 'X', 'Y', 'Z'))
    >> parse_one_theorem('congruent_triangle_property_angle_equal(RST,XYZ)')
    ('congruent_triangle_property_angle_equal', None, ('R', 'S', 'T', 'X', 'Y', 'Z'))
    >> parse_one_theorem('congruent_triangle_property_angle_equal')
    ('congruent_triangle_property_angle_equal', None, None)
    """
    if "(" not in theorem:
        return theorem, None, None

    t_name, t_para = theorem.split("(", 1)
    t_para = t_para.replace(")", "")
    if t_para[0].isnumeric():
        t_branch, t_para = t_para.split(",", 1)
        t_para = tuple(t_para.replace(",", ""))
        return t_name, t_branch, t_para

    t_para = tuple(t_para.replace(",", ""))
    return t_name, None, t_para



# Copyright (C) 2022-2023 FormalGeo Development Team
# Author: Xiaokai Zhang
# Contact: formalgeo@gmail.com

"""
'core' responsible for GPL statements executing, which consists of 2 submodules.
'GeometryPredicateLogicExecutor' responsible for GPL statements parsing and relational inference.
'EquationKiller' responsible for symbolic and algebraic computation.
"""

__all__ = [
    "GeometryPredicateLogicExecutor", "EquationKiller"
]

from formalgeo.core.engine import GeometryPredicateLogicExecutor, EquationKiller



# Copyright (C) 2022-2023 FormalGeo Development Team
# Author: Xiaokai Zhang
# Contact: formalgeo@gmail.com

"""
'tools' provides some practical tools, such as problem-solving process outputting.
"""

__all__ = [
    "load_json", "save_json", "safe_save_json", "debug_print", "rough_equal", "get_user_input",
    "simple_show", "show_solution", "get_solution_step", "get_solution_hypertree", "get_theorem_dag",
    "get_used_pid_and_theorem", "draw_solution_tree_and_theorem_dag"
]

from formalgeo.tools.utils import load_json, save_json, safe_save_json, debug_print, rough_equal, get_user_input
from formalgeo.tools.output import simple_show, show_solution, get_solution_step, get_solution_hypertree, get_theorem_dag
from formalgeo.tools.output import get_used_pid_and_theorem, draw_solution_tree_and_theorem_dag



# Copyright (C) 2022-2023 FormalGeo Development Team
# Author: Xiaokai Zhang
# Contact: formalgeo@gmail.com

"""
'parse' responsible for statements parsing and inverse parsing, building a bridge
between natural language, formal language, and machine language.
"""

__all__ = [
    "parse_expr", "get_expr_from_tree", "get_equation_from_tree",
    "parse_predicate_gdl", "parse_theorem_gdl", "parse_problem_cdl", "parse_theorem_seqs", "parse_one_theorem",
    "inverse_parse_one", "inverse_parse_logic_to_cdl", "inverse_parse_one_theorem"
]

from formalgeo.parse.basic import parse_expr, get_expr_from_tree, get_equation_from_tree
from formalgeo.parse.parse_tgdl import parse_theorem_gdl
from formalgeo.parse.parse_pgdl import parse_predicate_gdl
from formalgeo.parse.parse_cdl import parse_problem_cdl, parse_theorem_seqs, parse_one_theorem
from formalgeo.parse.inverse_parse_m2f import inverse_parse_one, inverse_parse_logic_to_cdl, inverse_parse_one_theorem



def inverse_parse_logic_to_cdl(problem):
    """Inverse parse conditions from machine language to formal language."""
    items = problem.condition.items
    steps = problem.condition.ids_of_step

    inverse_parsed_cdl = {}
    for step in sorted(list(steps)):
        if len(steps[step]) == 0:
            continue
        inverse_parsed_cdl[step] = []
        i = 0
        while i < len(steps[step]):
            _id = steps[step][i]
            predicate, item = items[_id][0], items[_id][1]

            if predicate in problem.parsed_predicate_GDL["Preset"]["Construction"] or \
                    predicate in problem.parsed_predicate_GDL["Preset"]["BasicEntity"]:  # skip preset
                i += 1
                continue

            result = inverse_parse_one(predicate, item, problem)
            inverse_parsed_cdl[step].append(result)

            if predicate in problem.parsed_predicate_GDL["Entity"]:  # remove duplicate representation
                i += len(problem.parsed_predicate_GDL["Entity"][predicate]["multi"]) + 1
            elif predicate in problem.parsed_predicate_GDL["Relation"]:
                i += len(problem.parsed_predicate_GDL["Relation"][predicate]["multi"]) + 1
            else:  # Equation
                i += 1

    return inverse_parsed_cdl


def inverse_parse_one(predicate, item, problem, remove_predicate=False):
    """
    Inverse parse one condition.
    Called by <inverse_parse_logic_to_cdl>.
    """
    if predicate == "Equation":
        inverse_parsed = inverse_parse_equation(item, problem)
    elif (predicate in problem.parsed_predicate_GDL["Preset"]["Construction"] or
          predicate in problem.parsed_predicate_GDL["Preset"]["BasicEntity"]):
        inverse_parsed = inverse_parse_preset(predicate, item)
    else:
        inverse_parsed = inverse_parse_logic(predicate, item, problem)

    if remove_predicate:
        inverse_parsed = inverse_parsed.split("(")[1].replace(")", "")

    return inverse_parsed


def inverse_parse_logic(predicate, item, problem):
    """
    Inverse parse logic conditions.
    >> inverse_parse_logic(Parallel, ('A', 'B', 'C', 'D'), problem)
    'Parallel(AB,CD)'
    """
    if predicate in problem.parsed_predicate_GDL["Entity"]:
        return predicate + "(" + "".join(item) + ")"
    else:
        result = []
        i = 0
        for l in problem.parsed_predicate_GDL["Relation"][predicate]["para_len"]:
            result.append("")
            for _ in range(l):
                result[-1] += item[i]
                i += 1
        return predicate + "(" + ",".join(result) + ")"


def inverse_parse_equation(item, problem):
    """
    Inverse parse algebra conditions.
    >> inverse_parse_one(ll_ac - ll_cd, problem)
    'Equal(LengthOfLine(AC),LengthOfLine(CD))'
    >> inverse_parse_one(ll_ac - 1, problem)
    'Value(LengthOfLine(AC))'
    >> inverse_parse_one(ll_ac - ll_cd - ll_ef, problem)
    'Equation(ll_ac-ll_cd-ll_ef)'
    """
    syms = list(item.free_symbols)
    if len(syms) == 1:
        if problem.condition.value_of_sym[syms[0]] is not None and \
                syms[0] - problem.condition.value_of_sym[syms[0]] == item:
            attr, items = problem.condition.attr_of_sym[syms[0]]
            if attr == "Free":
                attr = items[0][0]
            else:
                attr = attr + "(" + "".join(items[0]) + ")"
            return "Value({},{})".format(attr, str(problem.condition.value_of_sym[syms[0]]).replace(" ", ""))
    elif len(syms) == 2 and (item == (syms[0] - syms[1]) or item == (syms[1] - syms[0])):
        attr1, items1 = problem.condition.attr_of_sym[syms[0]]
        if attr1 == "Free":
            attr1 = items1[0][0]
        else:
            attr1 = attr1 + "(" + "".join(items1[0]) + ")"
        attr2, items2 = problem.condition.attr_of_sym[syms[1]]
        if attr2 == "Free":
            attr2 = items2[0][0]
        else:
            attr2 = attr2 + "(" + "".join(items2[0]) + ")"
        if item == (syms[0] - syms[1]):
            return "Equal({},{})".format(attr1, attr2)
        return "Equal({},{})".format(attr2, attr1)

    return "Equation" + "(" + str(item).replace(" ", "") + ")"


def inverse_parse_preset(predicate, item):
    """
    Inverse parse preset conditions.
    >> inverse_parse_logic(Line, ('A', 'B'))
    'Line(AB)'
    >> inverse_parse_logic(Cocircular, ('O', 'A', 'B', 'C'))
    'Cocircular(O,ABC)'
    """
    if predicate == "Cocircular":
        if len(item) == 1:
            return "Cocircular({})".format(item[0])
        else:
            return "Cocircular({},{})".format(item[0], "".join(item[1:]))
    elif predicate == "Shape":
        return "Shape({})".format(",".join(item))
    else:
        return "{}({})".format(predicate, "".join(item))


def inverse_parse_one_theorem(theorem, parsed_theorem_GDL):
    """
    Inverse parse theorem to formal language.
    >> inverse_parse_one_theorem('t_name', 1, ('R', 'S', 'T'), parsed_theorem_GDL)
    't_name(1,RST)'
    >> inverse_parse_one_theorem('t_name', None, ('R', 'S', 'T'), parsed_theorem_GDL)
    't_name(RST)'
    >> inverse_parse_one_theorem('t_name', None, None, parsed_theorem_GDL)
    't_name'
    """
    t_name, t_branch, t_para = theorem
    if t_para is None:
        if t_branch is None:
            return t_name
        return "{}({})".format(t_name, t_branch)
    else:
        result = []
        i = 0
        for l in parsed_theorem_GDL[t_name]["para_len"]:
            result.append("")
            for _ in range(l):
                result[-1] += t_para[i]
                i += 1
        t_para = ",".join(result)

        if t_branch is None:
            return "{}({})".format(t_name, t_para)
        return "{}({},{})".format(t_name, t_branch, t_para)



import copy
from formalgeo.parse import parse_expr, get_expr_from_tree, get_equation_from_tree


class Condition:
    def __init__(self):
        """All conditions of one problem."""
        self.id_count = 0  # <int>
        self.step_count = 0  # <int>
        self.fix_length_predicates = None  # <list> of <str>
        self.variable_length_predicates = None  # <list> of <str>

        self.items = []  # <list> of <tuple>, [(predicate, item, premise, theorem, step)]
        self.items_group = {}  # <dict>, [predicate: [item]], such as {'Angle':[('A', 'B', 'C')]}

        self.id_of_item = {}  # <dict>, {(predicate, item): id}, such as {('Angle', ('A', 'B', 'C')): 0}
        self.ids_of_predicate = {}  # <dict>, {predicate: [id]}, such as {'Angle': [0, 1, 2]}
        self.ids_of_step = {}  # <dict>, {step: [id]}, such as {0: [0, 1, 2]}

        self.sym_of_attr = {}  # <dict>, {(attr, paras): sym}, such as {('LengthOfLine', ('A', 'B')): l_ab}
        self.attr_of_sym = {}  # <dict>, {sym: (attr, (paras))}, such as {l_ab: ('LengthOfLine', (('A', 'B'),))}
        self.value_of_sym = {}  # <dict>, {sym: value}, such as {l_ab: 3}
        self.simplified_equation = {}  # <dict>, {simplified_equation: premises}, such as {a + b - 2: [1, 2, 3]}
        self.eq_solved = True  # <bool>, record whether the equation is solved

    def init_by_fl(self, fix_length_predicates, variable_length_predicates):
        """
        Initial condition by formal language.
        :param fix_length_predicates: List of predicates that has fix parameter length.
        :param variable_length_predicates: List of predicates that has variable parameter length.
        """
        self.fix_length_predicates = fix_length_predicates
        self.variable_length_predicates = variable_length_predicates
        self.ids_of_step[self.step_count] = []
        for predicate in self.fix_length_predicates + self.variable_length_predicates:
            self.items_group[predicate] = []
            self.ids_of_predicate[predicate] = []

    def init_by_copy(self, condition):
        """
        Initial condition by copy other condition.
        :param condition: <Condition>.
        """
        self.id_count = condition.id_count
        self.step_count = condition.step_count
        self.fix_length_predicates = copy.deepcopy(condition.fix_length_predicates)
        self.variable_length_predicates = copy.deepcopy(condition.variable_length_predicates)
        self.items = copy.deepcopy(condition.items)
        self.items_group = copy.deepcopy(condition.items_group)
        self.id_of_item = copy.deepcopy(condition.id_of_item)
        self.ids_of_predicate = copy.deepcopy(condition.ids_of_predicate)
        self.ids_of_step = copy.deepcopy(condition.ids_of_step)
        self.sym_of_attr = copy.deepcopy(condition.sym_of_attr)
        self.attr_of_sym = copy.deepcopy(condition.attr_of_sym)
        self.value_of_sym = copy.deepcopy(condition.value_of_sym)
        self.simplified_equation = copy.deepcopy(condition.simplified_equation)
        self.eq_solved = condition.eq_solved

    def add(self, predicate, item, premise, theorem):
        """
        Add one condition and guarantee no redundancy.
        :param predicate: <str>, predicate of condition.
        :param item: <tuple> of <str> or <equation>, body of condition, logic relation or equation.
        :param premise: <tuple> of <int>, premise of condition, no redundancy.
        :param theorem: <tuple>, (t_name, t_branch, t_para).
        :return added: <bool>, indicate whether the addition was successful.
        :return _id: <int>, id of condition, if not added, return None.
        """

        if not self.has(predicate, item):
            self.items.append((predicate, item, tuple(sorted(list(set(premise)))), theorem, self.step_count))
            self.items_group[predicate].append(item)
            self.ids_of_predicate[predicate].append(self.id_count)
            self.ids_of_step[self.step_count].append(self.id_count)

            if predicate == "Equation" and theorem[0] != "solve_eq":
                self.simplified_equation[item] = [self.id_count]
                self.eq_solved = False

            if predicate == "Equation":
                item = str(item)
            self.id_of_item[(predicate, item)] = self.id_count
            self.id_count += 1

            return True, self.id_count - 1

        return False, None

    def has(self, predicate, item):
        """
        Check if this condition exists.
        :param predicate: <str>, predicate of condition.
        :param item: <tuple> of <str> or symbols, body of condition, logic relation or equation.
        :return exist: <bool>, indicate whether the addition was successful.
        """
        if predicate == "Equation":
            return item in self.items_group["Equation"] or -item in self.items_group["Equation"]
        else:
            return item in self.items_group[predicate]

    def step(self):
        self.step_count += 1
        self.ids_of_step[self.step_count] = []

    def get_id_by_predicate_and_item(self, predicate, item):
        if predicate == "Equation":
            item = str(item)
        return self.id_of_item[(predicate, item)]

    def get_items_by_predicate(self, predicate):
        return copy.copy(self.items_group[predicate])

    def get_ids_and_items_by_predicate_and_variable(self, predicate, variable=None):
        ids = []
        items = []
        if variable is not None and predicate in self.variable_length_predicates:
            l = len(variable)
            for item in self.items_group[predicate]:
                if len(item) != l:
                    continue
                items.append(item)
                ids.append([self.id_of_item[(predicate, item)]])
        else:
            for item in self.items_group[predicate]:
                items.append(item)
                ids.append([self.id_of_item[(predicate, item)]])
        return ids, items

    def get_premise_by_predicate_and_item(self, predicate, item):
        return self.items[self.get_id_by_predicate_and_item(predicate, item)][2]

    def get_theorem_by_predicate_and_item(self, predicate, item):
        return self.items[self.get_id_by_predicate_and_item(predicate, item)][3]


class Goal:
    def __init__(self):
        """Goal of one problem."""
        self.type = None  # <str>, such as: 'algebra', 'logic'.
        self.item = None  # <equation> or predicate, such as: a - b, 'ParallelBetweenLine'
        self.answer = None  # <number> or <tuple> of <str>, such as: 0, ('A', 'B', 'C', 'D')
        self.solved = False  # <bool>
        self.solved_answer = None  # <number>, only used in 'algebra' and 'equal'
        self.premise = None  # <tuple> of <int>
        self.theorem = None  # <str>

    def init_by_fl(self, problem, goal_CDL):
        """Initial goal by formal language."""
        if goal_CDL["type"] == "value":
            self.type = "algebra"
            self.item = get_expr_from_tree(problem, goal_CDL["item"][1])
            self.answer = parse_expr(problem, goal_CDL["answer"])
        elif goal_CDL["type"] == "equal":
            self.type = "algebra"
            self.item = get_equation_from_tree(problem, goal_CDL["item"][1])
            self.answer = 0
        elif goal_CDL["type"] == "logic":
            self.type = "logic"
            self.item = goal_CDL["item"]
            self.answer = tuple(goal_CDL["answer"])

    def init_by_copy(self, goal):
        """Initial goal by copy."""
        self.type = goal.type
        self.item = goal.item
        self.answer = goal.answer
        self.solved = goal.solved
        self.solved_answer = goal.solved_answer
        self.premise = copy.copy(goal.premise)
        self.theorem = goal.theorem
